indexing
	description: "Program points: points within Eiffel classes, simple (correspond to single position in source code) or %
		%aggregative (correspong to set of positions), characterized by kinds of variables in scope %
		%and kind of assertions that are checked at it"
	author: "Nadia Polikarpova"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	CI_PROGRAM_POINT

inherit
	CI_CODE_GENERATION
		export {NONE}
			all
		end

	CI_OPERATOR_IDENTIFIERS
		export {NONE}
			all
		end

	CI_SETTINGS_CLIENT

	CI_PRINTABLE_CLASSES
		export {NONE}
			all
		end

feature -- Initialization
	make (a_class: like surrounding_class; a_universe: like universe) is
			-- Create a program point within `a_class' in `a_universe'
		require
			a_class_exists: a_class /= Void
			a_universe_exists: a_universe /= Void
		do
			surrounding_class := a_class
			universe := a_universe
			create {TWO_WAY_SORTED_SET [CI_VARIABLE]} variables.make
			variables.compare_objects
			add_variables
			create {LINKED_LIST [CI_INFERRED_ASSERTION]} inferred_assertions.make
			inferred_assertions.compare_objects
			create {LINKED_LIST [CI_SUPPRESSION]} suppressions.make
		ensure
			surrounding_class_set: surrounding_class = a_class
			universe_set: universe = a_universe
		end

feature -- Access
	surrounding_class: ET_CLASS
			-- Class to which program point belongs

	universe: ET_UNIVERSE
			-- Universe to which program point belongs			

	surrounding_routine: ET_ROUTINE
			-- Routine to which program point belongs (Void if it is a class-wide program point)

	surrounding_function: ET_FUNCTION
			-- Function to which program point belongs, if `is_function_point', Void otherwise
		do
			Result ?= surrounding_routine
		end

	kind (a_visitor: CI_PROGRAM_POINT_KIND_FORMAT): STRING is
			-- Program point kind name (for visiting)
		require
			a_visitor_exists: a_visitor /= Void
		deferred
		ensure
			result_exists: Result /= Void
		end

feature -- Status report
	ensures_invariant: BOOLEAN is
			-- Should invariant of instrumented class hold at program point?
		do
			-- By default answer in "No"
		end

	is_function_point: BOOLEAN is
			-- Does the point belong to a function?
		do
			-- By default answer in "No"	
		end

feature -- Variables
	variables: LINEAR_SUBSET [CI_VARIABLE]
			-- Variables in scope

	unfolded_variables: LIST [TREE [CI_VARIABLE]] is
			-- Unfolded variables (in a tree form)
		do
			if internal_unfolded_variables = Void then
				create internal_unfolded_variables.make (variables.count)
				from
					variables.start
				until
					variables.after
				loop
					internal_unfolded_variables.extend (variables.item.unfolded)
					variables.forth
				end
			end
			Result := internal_unfolded_variables
		ensure
			result_exists: Result /= Void
		end

	flattened_variables: LINEAR_SUBSET [CI_VARIABLE] is
			-- Flattened variables (unfolded variables flattened into a single set)
		do
			if internal_flattened_variables = Void then
				create {TWO_WAY_SORTED_SET [CI_VARIABLE]} internal_flattened_variables.make
				internal_flattened_variables.compare_objects
				from
					unfolded_variables.start
				until
					unfolded_variables.after
				loop
					flatten_variable (unfolded_variables.item)
					unfolded_variables.forth
				end
			end
			Result := internal_flattened_variables
		ensure
			result_exists: Result /= Void
		end

	named_variable (a_name: STRING): CI_VARIABLE is
			-- Variable with name `a_name' (from `flattened_variables')
		require
			a_name_exists: a_name /= Void
		do
			from
				flattened_variables.start
			until
				Result /= Void or flattened_variables.after
			loop
				if flattened_variables.item.name.is_equal (a_name) then
					Result := flattened_variables.item
				end
				flattened_variables.forth
			end
		end

feature {NONE} -- Variables
	add_variables is
			-- Find variables in scope and store them into `variables'
		require
			variables_exists: variables /= Void
		deferred
		end

	internal_unfolded_variables: ARRAYED_LIST [TREE [CI_VARIABLE]]
			-- Unfolded variables (in a tree form)

	internal_flattened_variables: LINEAR_SUBSET [CI_VARIABLE]
			-- Flattened variables (unfolded variables flattened into a single set)

	flatten_variable (an_unfolded_variable: TREE [CI_VARIABLE]) is
			-- Put all variables from `an_unfolded_variable' into `internal_flattened_variables'
		require
			an_unfolded_variable_exists: an_unfolded_variable /= Void
		do
			internal_flattened_variables.put (an_unfolded_variable.item)
			from
				an_unfolded_variable.child_start
			until
				an_unfolded_variable.child_after
			loop
				flatten_variable (an_unfolded_variable.child)
				an_unfolded_variable.child_forth
			end
		end

	add_constants is
			-- Find constant attributes of interest and add then to `variables'
		local
			c: ET_CLASS
			cursor: DS_HASH_SET_CURSOR [ET_CLASS]
		do
			add_constants_from (surrounding_class)
			if surrounding_class.suppliers /= Void then
				from
					cursor := surrounding_class.suppliers.new_cursor
					cursor.start
				until
					cursor.after
				loop
					c := cursor.item
					if c /= surrounding_class and Settings.has_interesting_constants (c) then
						add_constants_from (c)
					end
					cursor.forth
				end
			end
		end

	add_constants_from (a_class: ET_CLASS) is
			-- Add constant attributes defined in `a_class' to `variables'
		require
			a_class_exists: a_class /= Void
		local
			i: INTEGER
			queries: ET_QUERY_LIST
			query: ET_CONSTANT_ATTRIBUTE
			var: CI_CONSTANT
		do
			queries := a_class.queries
			from
				i := 1
			until
				i > queries.count
			loop
				query ?= queries.item (i)
				if query /= Void and Settings.is_valid_query_variable (query) and query.is_exported_to (surrounding_class, universe) then
					create {CI_CONSTANT} var.make (query, surrounding_routine, universe)
					if var.type.kind.is_string_reference then
						-- If it is a string constant, add its value, but not address
						variables.extend (create {CI_STRING_VALUE}.make (var))
					else
						variables.extend (var)
					end
				end
				i := i + 1
			end
		end

feature -- Assertions
	checked_assertions: ET_ASSERTIONS is
			-- Eiffel assertions, that are checked at this program point
		deferred
		end

	state_printing_function (printing_instructions: ET_COMPOUND; locals: ET_LOCAL_VARIABLE_LIST): CI_STATE_PRINTING_FUNCTION is
			-- Function to be called from `checked_assertions' in instrumented system, that prints visible state at this program point
			-- using `printing_instructions', needs `locals' as locals and returns True
		require
			printing_instructions_exists: printing_instructions /= Void
		local
			compound: ET_COMPOUND
		do
			create Result.make (state_printing_function_name, state_printing_function_formals, type_from_nongeneric_class (universe.boolean_class),
				surrounding_class)
			if locals /= Void and then not locals.is_empty then
				Result.set_locals (locals)
			end
			create compound.make_with_capacity (printing_instructions.count + 1)
			compound.put_first (create {ET_ASSIGNMENT}.make (tokens.result_keyword, tokens.true_keyword))
			insert_compound_to_front (compound, printing_instructions)
			Result.set_compound (compound)
		ensure
			result_exists: Result /= Void
		end

	instrument is
			-- Instrument program point by making `checked_assertions' call `state_printing_function'
		do
			create_assertions
--			checked_assertions.wipe_out
			checked_assertions.force_first (create {ET_CALL_EXPRESSION}.make (Void, state_printing_function_name, state_printing_function_actuals))
		end

	inferred_assertions: LIST [CI_INFERRED_ASSERTION]
			-- Assertions that were inferred for this program point

	annotate is
			-- Annotate program point with `inferred_assertions'
		do
			create_assertions
			process_original_assertions
			from
				inferred_assertions.finish
			until
				inferred_assertions.before
			loop
				checked_assertions.force_first (inferred_assertions.item)
				inferred_assertions.back
			end
		end

feature {NONE} -- Assertions
	create_assertions is
			-- Create empty `checked_assertions', if they don't exist
		deferred
		ensure
			checked_assertions_exists: checked_assertions /= Void
		end

	process_original_assertions is
			-- Process original assertions according to settings while annotating this point
		require
			checked_assertions_exists: checked_assertions /= Void
		deferred
		end

	state_printing_function_name: ET_FEATURE_NAME is
			-- Name of `state_printing_function' for this program point
		deferred
		end

	state_printing_function_formals: ET_FORMAL_ARGUMENT_LIST is
			-- Formal arguments of `state_printing_function'
		deferred
		end

	state_printing_function_actuals: ET_ACTUAL_ARGUMENT_LIST is
			-- Actual arguments of `state_printing_function' call inside instrumented `checked_assertions'
		deferred
		end

feature -- Suppressions
	suppressor: CI_PROGRAM_POINT is
			-- Program point, who's assertions suppress assertions at current program point (Void if no suppressor)
		obsolete "Use new suppression mechanism instead"
		do
			-- Default value is Void
		end

	suppressions: LIST [CI_SUPPRESSION]
			-- Suppressions for assertions at this program point (set by CI_PROGRAM_POINT_FINDER)

	clean_suppressions is
			-- Remove from `suppressions' those that don't have effect
		local
			has_effect:	BOOLEAN
		do
			from
				suppressions.start
			until
				suppressions.after
			loop
				has_effect := False
				from
					flattened_variables.start
				until
					flattened_variables.after or has_effect
				loop
					if flattened_variables.item.renamed (suppressions.item) /= Void then
						has_effect := True
					end
					flattened_variables.forth
				end
				if not has_effect then
					suppressions.remove
				else
					suppressions.forth
				end
			end
		end

invariant
	surrounding_class_exits: surrounding_class /= Void
	universe_exists: universe /= Void
	variables_exists: variables /= Void
	variables_compared_by_value: variables.object_comparison
	inferred_assertions_exists: inferred_assertions /= Void
	state_printing_function_name_exists: state_printing_function_name /= Void
	suppressions_exists: suppressions /= Void
end
