indexing
	description: "Objects that generate special classes that are responsible for runtime program state output."
	author: "Nadia Polikarpova"
	date: "$Date$"
	revision: "$Revision$"

class
	CI_RUNTIME_OUTPUT_GENERATOR

inherit
	CI_RUNTIME_OUTPUT_NAMES_CLIENT

	CI_SETTINGS_CLIENT

	CI_CODE_GENERATION
		export {NONE}
			all
		end

create
	make

feature -- Initialization
	make (a_file_name: like file_name; a_universe: like universe) is
			-- Create generator that associates trace file with `a_file_name' and creates classes within `a_universe'
		require
			a_file_name_valid: a_file_name /= Void and then a_file_name.is_valid
			a_universe_exists: a_universe /= Void
		do
			file_name := a_file_name
			universe := a_universe
		ensure
			file_name_set: file_name = a_file_name
			universe_set: universe = a_universe
		end

feature -- Basic operations
	generate_runtime_output is
			-- Generate "RUNTIME_OUTPUT" class
		local
			queries: ET_QUERY_LIST
			procedures: ET_PROCEDURE_LIST
		do
			RO_names.generate_runtime_output_name (universe)
			create runtime_output.make (RO_names.runtime_output_name, universe.classes.count + 1)
			runtime_output.set_group (Settings.instrumented_cluster)
			runtime_output.set_parsed
			universe.classes.force_new (runtime_output, runtime_output.name)

			runtime_output.set_feature_clauses (runtime_output_clauses)

			create queries.make
			create procedures.make
			generate_once_status_features (queries, procedures)
			queries.force_first (string_rep_feature)
			queries.force_first (file_feature)
			procedures.force_first (print_var_feature)
			runtime_output.set_queries (queries)
			runtime_output.set_procedures (procedures)
		ensure
			runtime_output_exists: runtime_output /= Void
			runtime_output_name_correct: runtime_output.name.same_class_name (RO_names.runtime_output_name)
			runtime_output_in_universe: universe.has_class (RO_names.runtime_output_name)
		end

	generate_runtime_output_client is
			-- Generate "RUNTIME_OUTPUT_CLIENT" class
		local
			parents: ET_PARENT_LIST
			queries: ET_QUERY_LIST
		do
			RO_names.generate_runtime_output_client_name (universe)
			if not Settings.is_any_changed then
				-- Generate separate "RUNTIME_OUTPUT_CLIENT"
				create runtime_output_client.make (RO_names.runtime_output_client_name, universe.classes.count + 1)
				runtime_output_client.set_group (Settings.instrumented_cluster)
				runtime_output_client.set_parsed
				universe.classes.force_new (runtime_output_client, runtime_output_client.name)

				runtime_output_client.set_class_mark (tokens.deferred_keyword)
				runtime_output_client.set_feature_clauses (runtime_output_client_clauses)

				create parents.make_with_capacity (1)
				parents.put_first (create {ET_PARENT}.make (universe.any_type, Void, Void, undefines, Void, Void))
				runtime_output_client.set_parents (parents)

				create queries.make_with_capacity (1)
				queries.put_first (trace_feature)
				runtime_output_client.set_queries (queries)
			else
				-- Do not generate "RUNTIME_OUTPUT_CLIENT", put its features into "ANY"
				runtime_output_client := universe.any_class
				runtime_output_client.queries.force_first (trace_feature)
				runtime_output_client.queries.first.set_feature_clause (runtime_output_client.feature_clauses.first)
			end
		ensure
			runtime_output_client_exists: runtime_output_client /= Void
			runtime_output_client_name_correct: runtime_output_client.name.same_class_name (RO_names.runtime_output_client_name)
			runtime_output_client_in_universe: universe.has_class (RO_names.runtime_output_client_name)
		end

feature -- Access
	runtime_output: ET_CLASS
			-- "RUNTIME_OUTPUT" class: singleton class, responsible for output and once status handling

	runtime_output_client: ET_CLASS
			-- Class that gives access to "RUNTIME_OUTPUT" singleton

feature {NONE} -- Features
	file_feature: ET_ONCE_FUNCTION is
			-- Query that yields output file
		local
			rescue_clause: ET_COMPOUND
		do
			-- Create function signature
			create Result.make (RO_names.file_feature_name, Void,
				type_from_nongeneric_class (universe.eiffel_class (tokens.plain_text_file_class_name)), runtime_output)

			-- Create locals
			Result.set_locals (single_local (create {ET_LOCAL_VARIABLE}.make (exceptions_local,
				type_from_nongeneric_class (universe.eiffel_class (tokens.exceptions_class_name)))))

			-- Create function body
			Result.set_compound (single_instruction (
				create {ET_CREATE_INSTRUCTION}.make (Void, tokens.result_keyword,
					create {ET_QUALIFIED_CALL}.make (tokens.make_open_write_feature_name,
					single_actual (manifest_string (file_name))))
				))
			Result.compound.set_keyword (tokens.once_keyword)

			create rescue_clause.make_with_capacity (3)
			rescue_clause.put_first (create {ET_CALL_INSTRUCTION}.make (exceptions_local, tokens.die_feature_name,
				single_actual (die_code)))
			rescue_clause.put_first (create {ET_CREATE_INSTRUCTION}.make (Void, exceptions_local, Void))
			rescue_clause.put_first (create {ET_CALL_INSTRUCTION}.make (Void, tokens.print_feature_name,
					single_actual (manifest_string (error_message))))
			Result.set_rescue_clause (rescue_clause)
			Result.rescue_clause.set_keyword (tokens.rescue_keyword)
			Result.set_feature_clause (output_clause)
		ensure
			result_exists: Result /= Void
		end

	print_var_feature: ET_DO_PROCEDURE is
			-- Procedure that prints variable info
		local
			args: ET_FORMAL_ARGUMENT_LIST
			compound: ET_COMPOUND
		do
			-- Create procedure signature
			create args.make_with_capacity (3)
			args.put_first (create {ET_FORMAL_ARGUMENT}.make (RO_names.modified_argument, universe.integer_type))
			args.put_first (create {ET_FORMAL_ARGUMENT}.make (RO_names.value_argument, type_from_nongeneric_class (universe.any_class)))
			args.put_first (create {ET_FORMAL_ARGUMENT}.make (RO_names.name_argument, universe.string_type))
			create Result.make (RO_names.print_var_feature_name, args, runtime_output)

			create compound.make_with_capacity (6)
			compound.put_first (new_line_instruction)
			compound.put_first (printing_instruction (RO_names.modified_argument))
			compound.put_first (new_line_instruction)
			compound.put_first (printing_instruction (RO_names.value_argument))
			compound.put_first (new_line_instruction)
			compound.put_first (printing_instruction (RO_names.name_argument))

			Result.set_compound (compound)
			Result.set_feature_clause (output_clause)
		ensure
			result_exists: Result /= Void
		end

	string_rep_feature: ET_DO_FUNCTION is
			-- Function that yields printable representation of string value
			-- (replaces all new lines by escape characters and adds double quotes)
		local
			compound: ET_COMPOUND
			args: ET_ACTUAL_ARGUMENT_LIST
			i: INTEGER
		do
			-- Create function signature
			create Result.make (RO_names.string_rep_feature_name,
				single_formal (create {ET_FORMAL_ARGUMENT}.make (RO_names.value_argument, universe.string_type)),
				universe.string_type,
				runtime_output)

			-- Create function body
			create compound.make_with_capacity (escape_sequences.count + 2)
			compound.put_first (create {ET_ASSIGNMENT}.make (tokens.result_keyword,
				create {ET_INFIX_EXPRESSION}.make (
						create {ET_INFIX_EXPRESSION}.make (
							tokens.double_quote,
							create {ET_SYMBOL_OPERATOR}.make_plus,
							tokens.result_keyword),
					create {ET_SYMBOL_OPERATOR}.make_plus, tokens.double_quote)
			))

			from
				i := escape_sequences.lower
			until
				i > escape_sequences.upper
			loop
				create args.make_with_capacity (2)
				args.put_first (double_escape_sequences.item (i))
				args.put_first (escape_sequences.item (i))
				compound.put_first (create {ET_CALL_INSTRUCTION}.make (tokens.result_keyword,
					tokens.replace_substring_all_feature_name, args))
				i := i + 1
			end

			compound.put_first (create {ET_ASSIGNMENT}.make (tokens.result_keyword,
				create {ET_CALL_EXPRESSION}.make (RO_names.value_argument, tokens.string_feature_name, Void)))

			Result.set_compound (compound)
			Result.set_feature_clause (output_clause)
		ensure
			result_exists: Result /= Void
		end

	trace_feature: ET_ONCE_FUNCTION is
			-- Function that yields an instance of "RUNTIME_OUTPUT"
		do
			create Result.make (RO_names.trace_feature_name, Void, type_from_nongeneric_class (runtime_output), runtime_output_client)
			Result.set_compound (single_instruction (create {ET_CREATE_INSTRUCTION}.make (Void, tokens.result_keyword, Void)))
			Result.compound.set_keyword (tokens.once_keyword)
			Result.set_feature_clause (access_clause)
		ensure
			result_exists: Result /= Void
		end

	called_feature (a_class_name: ET_CLASS_NAME; a_feature_name: ET_FEATURE_NAME): ET_ATTRIBUTE is
			-- Boolean query that reports, if feature called `a_feature_name' from class `a_class_name' has already been called
		do
			create Result.make (RO_names.called_feature_name (a_class_name, a_feature_name),
				type_from_nongeneric_class (universe.boolean_class), runtime_output)
			Result.set_feature_clause (once_status_clause)
		end

	set_called_feature (a_class_name: ET_CLASS_NAME; a_feature_name: ET_FEATURE_NAME): ET_DO_PROCEDURE is
			-- Procedure that sets corresponding `called_feature' for `a_class_name' and `a_feature_name' to `True'
		do
			create Result.make (RO_names.set_called_feature_name (a_class_name, a_feature_name), Void, runtime_output)
			Result.set_compound (single_instruction (
					create {ET_ASSIGNMENT}.make (RO_names.called_feature_name (a_class_name, a_feature_name), tokens.true_keyword)
				))
			Result.set_feature_clause (once_status_clause)
		end


feature {NONE} -- Feature clauses
	output_clause: ET_FEATURE_CLAUSE is
			-- "Output" clause in "RUNTIME_OUTPUT"
		once
			create Result.make (clients_any)
			add_comment ("Output", Result)
		ensure
			result_exists: Result /= Void
		end

	once_status_clause: ET_FEATURE_CLAUSE is
			-- "Once functions status" clause in "RUNTIME_OUTPUT"
		once
			create Result.make (clients_any)
			add_comment ("Once functions status", Result)
		ensure
			result_exists: Result /= Void
		end

	access_clause: ET_FEATURE_CLAUSE is
			-- "Access" clause in "RUNTIME_OUTPUT_CLIENT"
		once
			create Result.make (clients_any)
			add_comment ("Access", Result)
		ensure
			result_exists: Result /= Void
		end

	runtime_output_clauses: ET_FEATURE_CLAUSE_LIST is
			-- Feature clauses of "RUNTIME_OUTPUT"
		once
			create Result.make_with_capacity (2)
			Result.put_first (once_status_clause)
			Result.put_first (output_clause)
		ensure
			result_exists: Result /= Void
		end

	runtime_output_client_clauses: ET_FEATURE_CLAUSE_LIST is
			-- Feature clauses of "RUNTIME_OUTPUT_CLIENT"
		once
			create Result.make_with_capacity (1)
			Result.put_first (access_clause)
		ensure
			result_exists: Result /= Void
		end

feature {NONE} -- Implementation
	file_name: FILE_NAME
			-- Output file name

	universe: ET_UNIVERSE
			-- Syrrounding universe

	undefines: ET_KEYWORD_FEATURE_NAME_LIST is
			-- Undefines clause for "RUNTIME_OUTPUT_CLIENT"'s parent clause (all nonfrozen ANY features)
		local
			i: INTEGER
		do
			create Result.make_with_capacity (universe.any_class.queries.declared_count + universe.any_class.procedures.declared_count)
			Result.set_keyword (tokens.undefine_keyword)
			from
				i := universe.any_class.queries.declared_count
			until
				i < 1
			loop
				if not universe.any_class.queries.item (i).is_frozen then
					Result.put_first (universe.any_class.queries.item (i).name)
				end
				i := i - 1
			end
			from
				i := universe.any_class.procedures.declared_count
			until
				i < 1
			loop
				if not universe.any_class.procedures.item (i).is_frozen then
					Result.put_first (universe.any_class.procedures.item (i).name)
				end
				i := i - 1
			end
		end

	exceptions_local: ET_IDENTIFIER is
			-- Name of "exceptions" local variable in `trace_feature'
		once
			create Result.make ("exceptions")
		end

	die_code: ET_REGULAR_INTEGER_CONSTANT is
			-- Code to die with, if exception occures
		once
			create Result.make ("1")
		end

	error_message: STRING is "STATE PRINTING ERROR: Can't open trace file"
			-- Error messsage to be printed if exception occurs in `file_feature'

	file_expression: ET_EXPRESSION is
			-- Expression that yield output file from inside "RUNTIME_OUTPUT"
		once
			create {ET_CALL_EXPRESSION} Result.make (Void, RO_names.file_feature_name, Void)
		end

	printing_instruction (an_expression: ET_EXPRESSION) : ET_INSTRUCTION is
			-- An instruction that prints `an_expression' from inside "RUNTIME_OUTPUT"
		require
			an_expression_exists: an_expression /= Void
		local
			arg: ET_CALL_EXPRESSION
		do
			create arg.make (parenthesized_for_dot (an_expression), tokens.out_feature_name, Void)
			create {ET_CALL_INSTRUCTION} Result.make (file_expression, tokens.put_string_feature_name, single_actual (arg))
		ensure
			result_exists: Result /= Void
		end

	new_line_instruction: ET_INSTRUCTION is
			-- An instruction that prints a new line from inside "RUNTIME_OUTPUT"
		once
			create {ET_CALL_INSTRUCTION} Result.make (file_expression, tokens.new_line_feature_name, Void)
		end

	generate_once_status_features (a_queries: ET_QUERY_LIST; a_procedures: ET_PROCEDURE_LIST) is
			-- Fill `a_queries' with once status report queries and `a_procedures' with once status setting procedures
		require
			a_queries_exists: a_queries /= Void
			a_procedures_exists: a_procedures /= Void
		local
			c: ET_CLASS
			i: INTEGER
			query: ET_QUERY
		do
			from
				universe.classes.start
			until
				universe.classes.after
			loop
				c := universe.classes.item_for_iteration
				from
					i := c.queries.declared_count
				until
					i < 1
				loop
					query := c.queries.item (i)
					if Settings.is_valid_query_variable (query) and Settings.is_watched_once (query) then
						a_queries.force_first (called_feature (c.name, query.name))
						a_procedures.force_first (set_called_feature (c.name, query.name))
					end
					i := i - 1
				end
				universe.classes.forth
			end
		end

invariant
	file_name_exists: file_name /= Void
	file_name_valid: file_name.is_valid
	universe_exists: universe /= Void
	undefines_exists: undefines /= Void
	exceptions_local_exists: exceptions_local /= Void
	die_code_exists: die_code /= Void
	file_expression_exists: file_expression /= Void
	error_message_exists: error_message /= Void
	new_line_instruction_exists: new_line_instruction /= Void
end
