indexing
	description: "AST modifiers - processors that while traversing an AST build a new (modified) AST,%
		% in which some nodes can be replaced by nodes of other type.%
		% Instances of this class simply rebuild the same AST. To relace any nodes, inherit from this class%
		% and redefine any of `process_<node>' procedures.%
		% Note: At present doesn't process symbols."
	author: "Nadia Polikarpova"
	date: "$Date$"
	revision: "$Revision$"

class
	CI_AST_MODIFIER

inherit
	ET_AST_PROCESSOR

	CI_CODE_GENERATION
		export {NONE}
			all
		end

create
	make

feature -- Access
	modified_node: ET_AST_NODE
			-- Modified version of last fully processed node

feature {ET_AST_NODE} -- Processing

	process_actual_argument_list (a_list: ET_ACTUAL_ARGUMENT_LIST) is
			-- Replace each argument with its modified form
		local
			i, nb: INTEGER
			item: ET_EXPRESSION_ITEM
			res: ET_ACTUAL_ARGUMENT_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_actual_parameter_comma (a_parameter: ET_ACTUAL_PARAMETER_COMMA) is
			-- Replace `a_parameter' with its modified form
		local
			param: ET_ACTUAL_PARAMETER
		do
			good_type (a_parameter.actual_parameter, context, universe).process (Current)
			param ?= modified_node
			check param /= Void end
			create {ET_ACTUAL_PARAMETER_COMMA} modified_node.make (param, a_parameter.comma)
		end

	process_actual_parameter_list (a_list: ET_ACTUAL_PARAMETER_LIST) is
			-- Replace each parameter with its modified form
		local
			i, nb: INTEGER
			res: ET_ACTUAL_PARAMETER_LIST
			item: ET_ACTUAL_PARAMETER_ITEM
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				good_type (a_list.item (i), context, universe).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_agent_argument_operand_comma (an_argument: ET_AGENT_ARGUMENT_OPERAND_COMMA) is
			-- Replace actual argument with its modified form
		local
			arg: ET_AGENT_ARGUMENT_OPERAND
		do
			an_argument.agent_actual_argument.process (Current)
			arg ?= modified_node
			check arg /= Void end
			create {ET_AGENT_ARGUMENT_OPERAND_COMMA} modified_node.make (arg, an_argument.comma)
		end

	process_agent_argument_operand_list (a_list: ET_AGENT_ARGUMENT_OPERAND_LIST) is
			-- Replace each argument with its modified form
		local
			i, nb: INTEGER
			item: ET_AGENT_ARGUMENT_OPERAND_ITEM
			res: ET_AGENT_ARGUMENT_OPERAND_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_agent_implicit_current_target (a_target: ET_AGENT_IMPLICIT_CURRENT_TARGET) is
			-- Process `a_target'.
		do
			modified_node := a_target
		end

	process_agent_implicit_open_argument (an_argument: ET_AGENT_IMPLICIT_OPEN_ARGUMENT) is
			-- Process `an_argument'.
		do
			modified_node := an_argument
		end

	process_agent_keyword (a_keyword: ET_AGENT_KEYWORD) is
			-- Process `a_keyword'.
		do
			modified_node := a_keyword
		end

	process_agent_open_target (a_target: ET_AGENT_OPEN_TARGET) is
			-- Replace target type with its qualified form
		local
			t: ET_TYPE
		do
			good_type (a_target.type, context, universe).process (Current)
			t ?= modified_node
			check t /= Void end
			create {ET_AGENT_OPEN_TARGET} modified_node.make (t)
		end

	process_agent_typed_open_argument (an_argument: ET_AGENT_TYPED_OPEN_ARGUMENT) is
			-- Replace argument type with its qualified form
		local
			t: ET_TYPE
		do
			good_type (an_argument.type, context, universe).process (Current)
			t ?= modified_node
			check t /= Void end
			create {ET_AGENT_TYPED_OPEN_ARGUMENT} modified_node.make (t)
		end

	process_alias_free_name (a_name: ET_ALIAS_FREE_NAME) is
			-- Replace keyword and alias string with their modified form
		local
			key: ET_KEYWORD
			str: ET_MANIFEST_STRING
			res: ET_ALIAS_FREE_NAME
		do
			a_name.alias_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			a_name.alias_string.process (Current)
			str ?= modified_node
			check str /= Void end

			if a_name.is_prefix then
				create res.make_prefix (str)
			else
				create res.make_infix (str)
			end
			res.set_alias_keyword (key)

			modified_node := res
		end

	process_alias_name (a_name: ET_ALIAS_NAME) is
			-- Replace keyword and alias string with their modified form
		local
			str: ET_MANIFEST_STRING
			key: ET_KEYWORD
			res: ET_ALIAS_NAME
		do
			a_name.alias_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			a_name.alias_string.process (Current)
			str ?= modified_node
			check str /= Void end

			if str.value.is_equal (tokens.alias_and_name) then
				create res.make_and (str)
			elseif str.value.is_equal (tokens.alias_implies_name) then
				create res.make_implies (str)
			elseif str.value.is_equal (tokens.alias_or_name) then
				create res.make_or (str)
			elseif str.value.is_equal (tokens.alias_xor_name) then
				create res.make_xor (str)
			elseif str.value.is_equal (tokens.alias_div_name) then
				create res.make_div (str)
			elseif str.value.is_equal (tokens.alias_divide_name) then
				create res.make_divide (str)
			elseif str.value.is_equal (tokens.alias_ge_name) then
				create res.make_ge (str)
			elseif str.value.is_equal (tokens.alias_gt_name) then
				create res.make_gt (str)
			elseif str.value.is_equal (tokens.alias_le_name) then
				create res.make_le (str)
			elseif str.value.is_equal (tokens.alias_lt_name) then
				create res.make_lt (str)
			elseif str.value.is_equal (tokens.alias_minus_name) then
				create res.make_minus (str)
			elseif str.value.is_equal (tokens.alias_mod_name) then
				create res.make_mod (str)
			elseif str.value.is_equal (tokens.alias_plus_name) then
				create res.make_plus (str)
			elseif str.value.is_equal (tokens.alias_power_name) then
				create res.make_power (str)
			elseif str.value.is_equal (tokens.alias_times_name) then
				create res.make_times (str)
			elseif str.value.is_equal (tokens.alias_and_then_name) then
				create res.make_and_then (str)
			elseif str.value.is_equal (tokens.alias_or_else_name) then
				create res.make_or_else (str)
			elseif str.value.is_equal (tokens.alias_dotdot_name) then
				create res.make_dotdot (str)
			elseif str.value.is_equal (tokens.alias_not_name) then
				create res.make_not (str)
			elseif str.value.is_equal (tokens.alias_bracket_name) then
				create res.make_bracket (str)
			end
			res.set_alias_keyword (key)
			modified_node := res
		end

	process_aliased_feature_name (a_name: ET_ALIASED_FEATURE_NAME) is
			-- Replace feature name and alias with their modified form
		local
			feature_name: ET_IDENTIFIER
			alias_name: ET_ALIAS_NAME
		do
			a_name.feature_name.process (Current)
			feature_name ?= modified_node
			check feature_name /= Void end

			a_name.alias_name.process (Current)
			alias_name ?= modified_node
			check alias_name /= Void end

			create {ET_ALIASED_FEATURE_NAME} modified_node.make (feature_name, alias_name)
		end

	process_all_export (an_export: ET_ALL_EXPORT) is
			-- Perlace clients clause and keyword with their qialified form
		local
			clients: ET_CLIENTS
			key: ET_KEYWORD
			res: ET_ALL_EXPORT
		do
			an_export.clients_clause.process (Current)
			clients ?= modified_node
			check clients /= Void end

			an_export.all_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			create res.make (clients)
			res.set_all_keyword (key)
			modified_node := res
		end

	process_assertion_semicolon (an_assertion: ET_ASSERTION_SEMICOLON) is
			-- Replace assertion with its modified form
		local
			assertion: ET_ASSERTION
		do
			an_assertion.assertion.process (Current)
			assertion ?= modified_node
			check assertion /= Void end

			create {ET_ASSERTION_SEMICOLON} modified_node.make (assertion, an_assertion.semicolon)
		end

	process_assign_feature_name (an_assigner: ET_ASSIGN_FEATURE_NAME) is
			-- Perlace keyword and feature name with their modified form
		local
			feature_name: ET_FEATURE_NAME
			key: ET_KEYWORD
			res: ET_ASSIGN_FEATURE_NAME
		do
			an_assigner.assign_keyword.process (Current)
			feature_name ?= modified_node
			check feature_name /= Void end

			an_assigner.feature_name.process (Current)
			key ?= modified_node
			check key /= Void end

			create res.make (feature_name)
			res.set_assign_keyword (key)
			modified_node := res
		end

	process_assigner_instruction (an_instruction: ET_ASSIGNER_INSTRUCTION) is
			-- Replace call and source with their modified form
		local
			call: ET_FEATURE_CALL_EXPRESSION
			source: ET_EXPRESSION
		do
			an_instruction.call.process (Current)
			call ?= modified_node
			check call /= Void end

			an_instruction.source.process (Current)
			source ?= modified_node
			check source /= Void end

			create {ET_ASSIGNER_INSTRUCTION} modified_node.make (call, source)
		end

	process_assignment (an_instruction: ET_ASSIGNMENT) is
			-- Replace target and source with their modified form
		local
			target: ET_WRITABLE
			source: ET_EXPRESSION
		do
			an_instruction.target.process (Current)
			target ?= modified_node
			check target /= Void end

			an_instruction.source.process (Current)
			source ?= modified_node
			check source /= Void end

			create {ET_ASSIGNMENT} modified_node.make (target, source)
		end

	process_assignment_attempt (an_instruction: ET_ASSIGNMENT_ATTEMPT) is
			-- Replace target and source with their modified form
		local
			target: ET_WRITABLE
			source: ET_EXPRESSION
		do
			an_instruction.target.process (Current)
			target ?= modified_node
			check target /= Void end

			an_instruction.source.process (Current)
			source ?= modified_node
			check source /= Void end

			create {ET_ASSIGNMENT_ATTEMPT} modified_node.make (target, source)
		end

	process_attribute (a_feature: ET_ATTRIBUTE) is
			-- Replace frozen keyword, name, type and assigner with their modified form
		local
			frozen_key: ET_KEYWORD
			type: ET_DECLARED_TYPE
			extended_name: ET_EXTENDED_FEATURE_NAME
			assigner: ET_ASSIGNER
			res: ET_ATTRIBUTE
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end
			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			good_type (a_feature.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			create res.make (extended_name, type, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)

			assigner := a_feature.assigner
			if assigner /= Void then
				assigner.process (Current)
				assigner ?= modified_node
				check modified_node /= Void implies assigner /= Void end
			end
			res.set_assigner (assigner)

			modified_node := res
		end

	process_bang_instruction (an_instruction: ET_BANG_INSTRUCTION) is
			-- Replace type, target and call with their modified form
		local
			type: ET_TYPE
			target: ET_WRITABLE
			call: ET_QUALIFIED_CALL
		do
			type ?= good_type (an_instruction.type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end

			an_instruction.target.process (Current)
			target ?= modified_node
			check target /= Void end

			call := an_instruction.creation_call
			if call /= Void then
				call.process (Current)
				call ?= modified_node
				check modified_node /= Void implies call /= Void end
			end

			create {ET_BANG_INSTRUCTION} modified_node.make (type, target, call)
		end

	process_bit_constant (a_constant: ET_BIT_CONSTANT) is
			-- Process `a_constant'.
		do
			modified_node := a_constant
		end

	process_bit_feature (a_type: ET_BIT_FEATURE) is
			-- Replace keyword and feature name with their modified form
		local
			key: ET_IDENTIFIER
			name: ET_IDENTIFIER
			res: ET_BIT_FEATURE
		do
			a_type.bit_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			a_type.name.process (Current)
			name ?= modified_node
			check name /= Void end

			create res.make (name)
			res.set_bit_keyword (key)
			modified_node := res
		end

	process_bit_n (a_type: ET_BIT_N) is
			-- Replace ketword and constant with their modified form
		local
			key: ET_IDENTIFIER
			constant: ET_INTEGER_CONSTANT
			res: ET_BIT_N
		do
			a_type.bit_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			a_type.constant.process (Current)
			constant ?= modified_node
			check constant /= Void end

			create res.make (constant)
			res.set_bit_keyword (key)
			modified_node := res
		end

	process_braced_class_name (a_name: ET_BRACED_CLASS_NAME) is
			-- Replace class name with its modified form
		local
			class_name: ET_CLASS_NAME
		do
			a_name.class_name.process (Current)
			class_name ?= modified_node
			check class_name /= Void end
			create {ET_BRACED_CLASS_NAME} modified_node.make (class_name)
		end

	process_braced_type (a_type: ET_BRACED_TYPE) is
			-- Replace type with its modified form
		local
			type: ET_TYPE
		do
			good_type (a_type.type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end
			create {ET_BRACED_TYPE} modified_node.make (type)
		end

	process_braced_type_list (a_list: ET_BRACED_TYPE_LIST) is
			-- Replace each type with its qualified form
		local
			i, nb: INTEGER
			item: ET_TYPE_ITEM
			res: ET_BRACED_TYPE_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				good_type (a_list.item (i), context, universe).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_bracket_argument_list (a_list: ET_BRACKET_ARGUMENT_LIST) is
			-- Replace each argument with its qualified form
		local
			i, nb: INTEGER
			item: ET_EXPRESSION_ITEM
			res: ET_BRACKET_ARGUMENT_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_bracket_expression (an_expression: ET_BRACKET_EXPRESSION) is
			-- Replace target and arguments with their modified form
		local
			t: ET_EXPRESSION
			args: ET_BRACKET_ARGUMENT_LIST
		do
			t := an_expression.target
			t.process (Current)
			t ?= modified_node
			check t /= Void end

			args := an_expression.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			create {ET_BRACKET_EXPRESSION} modified_node.make (t, an_expression.name, args)
		end

	process_bracket_symbol (a_symbol: ET_BRACKET_SYMBOL) is
			-- Process `a_symbol'.
		do
			modified_node := a_symbol
		end

	process_break (a_break: ET_BREAK) is
			-- Process `a_break'.
		do
		end

	process_c1_character_constant (a_constant: ET_C1_CHARACTER_CONSTANT) is
			-- Replace cast type with its modified form
		local
			type: ET_TARGET_TYPE
			res: ET_C1_CHARACTER_CONSTANT
		do
			type ?= good_type (a_constant.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'

			create res.make (a_constant.value)
			res.set_cast_type (type)
			modified_node := res
		end

	process_c2_character_constant (a_constant: ET_C2_CHARACTER_CONSTANT) is
			-- Replace cast type with its modified form
		local
			type: ET_TARGET_TYPE
			res: ET_C2_CHARACTER_CONSTANT
		do
			type ?= good_type (a_constant.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'

			create res.make (a_constant.value)
			res.set_cast_type (type)
			modified_node := res
		end

	process_c3_character_constant (a_constant: ET_C3_CHARACTER_CONSTANT) is
			-- Replace cast type with its modified form
		local
			type: ET_TARGET_TYPE
			res: ET_C3_CHARACTER_CONSTANT
		do
			type ?= good_type (a_constant.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'

			create res.make (a_constant.literal, a_constant.value)
			res.set_cast_type (type)
			modified_node := res
		end

	process_call_agent (an_expression: ET_CALL_AGENT) is
			-- Replace keyword, target, name and arguments with their modified form
		local
			key: ET_KEYWORD
			target: ET_AGENT_TARGET
			qualified_name: ET_QUALIFIED_FEATURE_NAME
			args: ET_AGENT_ARGUMENT_OPERAND_LIST
			res: ET_CALL_AGENT
		do
			an_expression.agent_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			an_expression.target.process (Current)
			target ?= modified_node
			check target /= Void end

			an_expression.qualified_name.process (Current)
			qualified_name ?= modified_node
			check qualified_name /= Void end

			args ?= an_expression.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			create res.make (target, qualified_name, args)
			res.set_agent_keyword (key)
			modified_node := res
		end

	process_call_expression (an_expression: ET_CALL_EXPRESSION) is
			-- Replace target, name and arguments with their modified form
		local
			target: ET_EXPRESSION
			qualified_name: ET_QUALIFIED_FEATURE_NAME
			args: ET_ACTUAL_ARGUMENT_LIST
		do
			target := an_expression.target
			if target /= Void then
				target.process (Current)
				target ?= modified_node
				check modified_node /= Void implies target /= Void end
			end

			an_expression.qualified_name.process (Current)
			qualified_name ?= modified_node
			check qualified_name /= Void end

			args := an_expression.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			create {ET_CALL_EXPRESSION} modified_node.make (target, qualified_name, args)
		end

	process_call_instruction (an_instruction: ET_CALL_INSTRUCTION) is
			-- Replace target, name and arguments with their modified form
		local
			target: ET_EXPRESSION
			qualified_name: ET_QUALIFIED_FEATURE_NAME
			args: ET_ACTUAL_ARGUMENT_LIST
		do
			target := an_instruction.target
			if target /= Void then
				target.process (Current)
				target ?= modified_node
				check modified_node /= Void implies target /= Void end
			end

			an_instruction.qualified_name.process (Current)
			qualified_name ?= modified_node
			check qualified_name /= Void end

			args := an_instruction.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			create {ET_CALL_INSTRUCTION} modified_node.make (target, qualified_name, args)
		end

	process_check_instruction (an_instruction: ET_CHECK_INSTRUCTION) is
			-- Replace keywords and each assertion with their modified form
		local
			i, nb: INTEGER
			item: ET_ASSERTION_ITEM
			res: ET_CHECK_INSTRUCTION
			check_key, end_key: ET_KEYWORD
		do
			an_instruction.check_keyword.process (Current)
			check_key ?= modified_node
			check check_key /= Void end

			nb := an_instruction.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				an_instruction.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end

			an_instruction.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			res.set_check_keyword (check_key)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_choice_comma (a_choice: ET_CHOICE_COMMA) is
			-- Replace choice with its modified form
		local
			choice: ET_CHOICE
		do
			a_choice.choice.process (Current)
			choice ?= modified_node
			check choice /= Void end

			create {ET_CHOICE_COMMA} modified_node.make (choice, a_choice.comma)
		end

	process_choice_list (a_list: ET_CHOICE_LIST) is
			-- Replace when keyword and each choice item with its modified form
		local
			when_key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_CHOICE_ITEM
			res: ET_CHOICE_LIST
		do
			a_list.when_keyword.process (Current)
			when_key ?= modified_node
			check when_key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_when_keyword (when_key)

			modified_node := res
		end

	process_choice_range (a_choice: ET_CHOICE_RANGE) is
			-- Replace lower and upper with their modified form
		local
			lower, upper: ET_CHOICE_CONSTANT
		do
			a_choice.lower.process (Current)
			lower ?= modified_node
			check lower /= Void end

			a_choice.upper.process (Current)
			upper ?= modified_node
			check upper /= Void end

			create {ET_CHOICE_RANGE} modified_node.make (lower, upper)
		end

	process_class (a_class: ET_CLASS) is
			-- Replace all class components with their modified form
		local
			frozen_key, class_key, class_mark, external_key, end_key: ET_KEYWORD
			index1, index2: ET_INDEXING_LIST
			formals: ET_FORMAL_PARAMETER_LIST
			obsolete_message: ET_OBSOLETE
			class_name: ET_CLASS_NAME
			parents: ET_PARENT_LIST
			creators: ET_CREATOR_LIST
			convert_features: ET_CONVERT_FEATURE_LIST
			feature_clauses: ET_FEATURE_CLAUSE_LIST
			feature_clause: ET_FEATURE_CLAUSE
			queries: ET_QUERY_LIST
			query: ET_QUERY
			procedures: ET_PROCEDURE_LIST
			procedure: ET_PROCEDURE
			nb, i, clause_num: INTEGER
			invariants: ET_INVARIANTS
			res: ET_CLASS
		do
			index1 := a_class.first_indexing
			if index1 /= Void then
				index1.process (Current)
				index1 ?= modified_node
				check modified_node /= Void implies index1 /= Void end
			end

			frozen_key := a_class.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			class_mark := a_class.class_mark
			if class_mark /= Void then
				class_mark.process (Current)
				class_mark ?= modified_node
				check modified_node /= Void implies class_mark /= Void end
			end

			external_key := a_class.external_keyword
			if external_key /= Void then
				external_key.process (Current)
				external_key ?= modified_node
				check modified_node /= Void implies external_key /= Void end
			end

			a_class.class_keyword.process (Current)
			class_key ?= modified_node
			check class_key /= Void end

			a_class.name.process (Current)
			class_name ?= modified_node
			check class_name /= Void end

			formals := a_class.formal_parameters
			if formals /= Void then
				formals.process (Current)
				formals ?= modified_node
				check modified_node /= Void implies formals /= Void end
			end

			obsolete_message := a_class.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			parents := a_class.parents
			if parents /= Void then
				parents.process (Current)
				parents ?= modified_node
				check modified_node /= Void implies parents /= Void end
			end

			creators := a_class.creators
			if creators /= Void then
				creators.process (Current)
				creators ?= modified_node
				check modified_node /= Void implies creators /= Void end
			end

			convert_features := a_class.convert_features
			if convert_features /= Void then
				convert_features.process (Current)
				convert_features ?= modified_node
				check modified_node /= Void implies convert_features /= Void end
			end

			nb := a_class.feature_clauses.count
			create feature_clauses.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_class.feature_clauses.item (i).process (Current)
				feature_clause ?= modified_node
				check feature_clause /= Void end
				feature_clauses.put_first (feature_clause)
				i := i - 1
			end

			nb := a_class.queries.count
			create queries.make_with_capacity (nb)
			clause_num := feature_clauses.count
			from i := nb until i < 1 loop
				a_class.queries.item (i).process (Current)
				query ?= modified_node
				check query /= Void end
				if query.implementation_class = a_class then
					from until a_class.feature_clauses.item (clause_num) = a_class.queries.item(i).feature_clause
					loop clause_num := clause_num - 1 end
					query.set_feature_clause (feature_clauses.item (clause_num))
				end
				queries.put_first (query)
				i := i - 1
			end

			nb := a_class.procedures.count
			create procedures.make_with_capacity (nb)
			clause_num := feature_clauses.count
			from i := nb until i < 1 loop
				a_class.procedures.item (i).process (Current)
				procedure ?= modified_node
				check procedure /= Void end
				if procedure.implementation_class = a_class then
					from until a_class.feature_clauses.item (clause_num) = a_class.procedures.item (i).feature_clause
					loop clause_num := clause_num - 1 end
					procedure.set_feature_clause (feature_clauses.item (clause_num))
				end
				procedures.put_first (procedure)
				i := i - 1
			end

			invariants := a_class.invariants
			if invariants /= Void then
				invariants.process (Current)
				invariants ?= modified_node
				check modified_node /= Void implies invariants /= Void end
			end

			index2 := a_class.second_indexing
			if index2 /= Void then
				index2.process (Current)
				index2 ?= modified_node
				check modified_node /= Void implies index2 /= Void end
			end

			a_class.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (class_name, a_class.id)
			res.set_first_indexing (index1)
			res.set_frozen_keyword (frozen_key)
			res.set_class_mark (class_mark)
			res.set_external_keyword (external_key)
			res.set_class_keyword (class_key)
			res.set_formal_parameters (formals)
			res.set_obsolete_message (obsolete_message)
			res.set_parents (parents)
			res.set_creators (creators)
			res.set_convert_features (convert_features)
			res.set_feature_clauses (feature_clauses)
			res.set_queries (queries)
			res.set_procedures (procedures)
			res.set_invariants (invariants)
			res.set_second_indexing (index2)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_class_name_comma (a_name: ET_CLASS_NAME_COMMA) is
			-- Replace class name with its modified form
		local
			class_name: ET_CLASS_NAME
		do
			a_name.class_name.process (Current)
			class_name ?= modified_node
			check class_name /= Void end
			create {ET_CLASS_NAME_COMMA} modified_node.make (class_name, a_name.comma)
		end

	process_class_name_list (a_list: ET_CLASS_NAME_LIST) is
			-- Replcae each class name with its modified form
		local
			i, nb: INTEGER
			item: ET_CLASS_NAME_ITEM
			res: ET_CLIENTS
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_class_type (a_type: ET_CLASS_TYPE) is
			-- Replace type mark and name with their modified form
		local
			type_mark: ET_KEYWORD
			name: ET_CLASS_NAME
		do
			type_mark := a_type.type_mark
			if type_mark /= Void then
				type_mark.process (Current)
				type_mark ?= modified_node
				check modified_node /= Void implies type_mark /= Void end
			end

			a_type.name.process (Current)
			name ?= modified_node
			check name /= Void end

			create {ET_CLASS_TYPE} modified_node.make (type_mark, name, universe.eiffel_class (name))
		end

	process_clients (a_list: ET_CLIENTS) is
			-- Replcae each client with its modified form
		do
			process_class_name_list (a_list)
		end

	process_colon_type (a_type: ET_COLON_TYPE) is
			-- Replace type with its modified form
		local
			type: ET_TYPE
		do
			good_type (a_type.type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end
			create {ET_COLON_TYPE} modified_node.make (a_type.colon, type)
		end

	process_compound (a_list: ET_COMPOUND) is
			-- Replace keyword and each instruction with its modified form
		local
			key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_INSTRUCTION
			res: ET_COMPOUND
		do
			a_list.keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_keyword (key)
			modified_node := res
		end

	process_constant_attribute (a_feature: ET_CONSTANT_ATTRIBUTE) is
			-- Replace keywords, name, type, assigner and constant with their modified form
		local
			frozen_key, is_key: ET_KEYWORD
			type: ET_DECLARED_TYPE
			extended_name: ET_EXTENDED_FEATURE_NAME
			constant: ET_CONSTANT
			assigner: ET_ASSIGNER
			res: ET_CONSTANT_ATTRIBUTE
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end
			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			good_type (a_feature.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			assigner := a_feature.assigner
			if assigner /= Void then
				assigner.process (Current)
				assigner ?= modified_node
				check modified_node /= Void implies assigner /= Void end
			end

			a_feature.is_keyword.process (Current)
			is_key ?= modified_node
			check is_key /= Void end

			a_feature.constant.process (Current)
			constant ?= modified_node
			check constant /= Void end

			create res.make (extended_name, type, constant, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_is_keyword (is_key)
			res.set_assigner (assigner)

			modified_node := res
		end

	process_constrained_formal_parameter (a_parameter: ET_CONSTRAINED_FORMAL_PARAMETER) is
			-- Replace name, constraint, creation procedures and type mark with their modified form
		local
			name: ET_IDENTIFIER
			constraint: ET_TYPE
			creation_procedures: ET_CONSTRAINT_CREATOR
			type_mark: ET_KEYWORD
			res: ET_CONSTRAINED_FORMAL_PARAMETER
		do
			type_mark := a_parameter.type_mark
			if type_mark /= Void then
				type_mark.process (Current)
				type_mark ?= modified_node
				check modified_node /= Void implies type_mark /= Void end
			end

			a_parameter.name.process (Current)
			name ?= modified_node
			check name /= Void end

			good_type (a_parameter.constraint, context, universe).process (Current)
			constraint ?= modified_node
			check constraint /= Void end

			creation_procedures := a_parameter.creation_procedures
			if creation_procedures /= Void then
				creation_procedures.process (Current)
				creation_procedures ?= modified_node
				check modified_node /= Void implies creation_procedures /= Void end
			end

			create res.make (name, constraint, creation_procedures)
			res.set_type_mark (type_mark)
			modified_node := res
		end

	process_constraint_creator (a_list: ET_CONSTRAINT_CREATOR) is
			-- Replace keywords and each feature name with its modified form
		local
			create_key, end_key: ET_KEYWORD
			item: ET_FEATURE_NAME_ITEM
			i, nb: INTEGER
			res: ET_CONSTRAINT_CREATOR
		do
			a_list.create_keyword.process (Current)
			create_key ?= modified_node
			check create_key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end

			a_list.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			res.set_create_keyword (create_key)
			res.set_end_keyword (end_key)
		end

	process_convert_expression (a_convert_expression: ET_CONVERT_EXPRESSION) is
			-- Replace expression and convert feature with their modified form
		local
			expr: ET_EXPRESSION
--			convert_feature: ET_CONVERT_FEATURE
		do
			a_convert_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end

--			a_convert_expression.convert_feature.process (Current)
--			convert_feature ?= modified_node
--			check convert_feature /= Void end

			create {ET_CONVERT_EXPRESSION} modified_node.make (expr, a_convert_expression.convert_feature)
		end

	process_convert_feature_comma (a_convert_feature: ET_CONVERT_FEATURE_COMMA) is
			-- Replace feature with its modified form
		local
			convert_feature: ET_CONVERT_FEATURE
		do
			a_convert_feature.convert_feature.process (Current)
			convert_feature ?= modified_node
			check convert_feature /= Void end

			create {ET_CONVERT_FEATURE_COMMA} modified_node.make (convert_feature, a_convert_feature.comma)
		end

	process_convert_feature_list (a_list: ET_CONVERT_FEATURE_LIST) is
			-- Replace keyword and each feature with its modified- form
		local
			key: ET_KEYWORD
			item: ET_CONVERT_FEATURE_ITEM
			i, nb: INTEGER
			res: ET_CONVERT_FEATURE_LIST
		do
			a_list.convert_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_convert_keyword (key)
			modified_node := res
		end

	process_convert_function (a_convert_function: ET_CONVERT_FUNCTION) is
			-- Replace name and type list with their modified form
		local
			name: ET_FEATURE_NAME
			list: ET_TYPE_LIST
		do
			a_convert_function.name.process (Current)
			name ?= modified_node
			check name /= Void end

			a_convert_function.types.process (Current)
			list ?= modified_node
			check list /= Void end

			create {ET_CONVERT_FUNCTION} modified_node.make (name, list)
		end

	process_convert_procedure (a_convert_procedure: ET_CONVERT_PROCEDURE) is
			-- Replace name and type list with their modified form
		local
			name: ET_FEATURE_NAME
			list: ET_TYPE_LIST
		do
			a_convert_procedure.name.process (Current)
			name ?= modified_node
			check name /= Void end

			a_convert_procedure.types.process (Current)
			list ?= modified_node
			check list /= Void end

			create {ET_CONVERT_PROCEDURE} modified_node.make (name, list)
		end

	process_convert_to_expression (a_convert_expression: ET_CONVERT_TO_EXPRESSION) is
			-- Replace expression and convert feature with their modified form
		local
			expr: ET_EXPRESSION
			convert_feature: ET_CONVERT_FEATURE
		do
			a_convert_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end

			a_convert_expression.convert_feature.process (Current)
			convert_feature ?= modified_node
			check convert_feature /= Void end

			create {ET_CONVERT_TO_EXPRESSION} modified_node.make (expr, convert_feature)
		end

	process_create_expression (an_expression: ET_CREATE_EXPRESSION) is
			-- Replace keyword, creation type and call by their modified form
		local
			key: ET_KEYWORD
			type: ET_TARGET_TYPE
			call: ET_QUALIFIED_CALL
			res: ET_CREATE_EXPRESSION
		do
			an_expression.create_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			good_type (an_expression.creation_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			call := an_expression.creation_call
			if call /= Void then
				call.process (Current)
				call ?= modified_node
				check modified_node /= Void implies call /= Void end
			end

			create res.make (type, call)
			res.set_create_keyword (key)
			modified_node := res
		end

	process_create_instruction (an_instruction: ET_CREATE_INSTRUCTION) is
			-- Replace keyword, creation type and call by their modified form
		local
			key: ET_KEYWORD
			type: ET_TARGET_TYPE
			target: ET_WRITABLE
			call: ET_QUALIFIED_CALL
			res: ET_CREATE_INSTRUCTION
		do
			an_instruction.create_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			type ?= good_type (an_instruction.creation_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end

			an_instruction.target.process (Current)
			target ?= modified_node
			check target /= Void end

			call := an_instruction.creation_call
			if call /= Void then
				call.process (Current)
				call ?= modified_node
				check modified_node /= Void implies call /= Void end
			end

			create res.make (type, target, call)
			res.set_create_keyword (key)
			modified_node := res
		end

	process_creator (a_list: ET_CREATOR) is
			-- Replace keyword, clients and each feature name with its modified form
		local
			key: ET_KEYWORD
			clients: ET_CLASS_NAME_LIST	-- Should be of type ET_CLIENTS
			item: ET_FEATURE_NAME_ITEM
			i, nb: INTEGER
			res: ET_CREATOR
		do
			a_list.creation_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			clients := a_list.clients -- a_list.client_clause: problem, is Void
			if clients /= Void then
				process_class_name_list (clients) -- clients.process (Current)
				clients ?= modified_node
				check modified_node /= Void implies clients /= Void end
			end

			nb := a_list.count
			create res.make_with_capacity (clients, nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_creation_keyword (key)
			modified_node := res
		end

	process_creator_list (a_list: ET_CREATOR_LIST) is
			-- Replace each creator with its modified form
		local
			item: ET_CREATOR
			i, nb: INTEGER
			res: ET_CREATOR_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_current (an_expression: ET_CURRENT) is
			-- Process `an_expression'.
		do
			modified_node := an_expression
		end

	process_current_address (an_expression: ET_CURRENT_ADDRESS) is
			-- Replace "current" with its modified form
		local
			cur: ET_CURRENT
		do
			an_expression.current_keyword.process (Current)
			cur ?= modified_node
			check cur /= Void end
		end

	process_custom_attribute (an_attribute: ET_CUSTOM_ATTRIBUTE) is
			-- Process `an_attribute'.
		do
			-- TODO: support .NET
			modified_node := an_attribute
		end

	process_debug_instruction (an_instruction: ET_DEBUG_INSTRUCTION) is
			-- Process `an_instruction'.
		local
			compound: ET_COMPOUND
			keys: ET_MANIFEST_STRING_LIST
			end_key: ET_KEYWORD
			res: ET_DEBUG_INSTRUCTION
		do
			compound := an_instruction.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			keys := an_instruction.keys
			if keys /= Void then
				keys.process (Current)
				keys ?= modified_node
				check modified_node /= Void implies keys /= Void end
			end

			an_instruction.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (keys, compound)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_deferred_function (a_feature: ET_DEFERRED_FUNCTION) is
			-- Replace all feature components with their modified form
		local
			frozen_key, is_key, deferred_key, end_key: ET_KEYWORD
			extended_name: ET_EXTENDED_FEATURE_NAME
			args: ET_FORMAL_ARGUMENT_LIST
			type: ET_DECLARED_TYPE
			index: ET_INDEXING_LIST
			obsolete_message: ET_OBSOLETE
			preconditions: ET_PRECONDITIONS
			postconditions: ET_POSTCONDITIONS
			assigner: ET_ASSIGNER
			res: ET_DEFERRED_FUNCTION
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			args := a_feature.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			good_type (a_feature.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			assigner := a_feature.assigner
			if assigner /= Void then
				assigner.process (Current)
				assigner ?= modified_node
				check modified_node /= Void implies assigner /= Void end
			end

			is_key := a_feature.is_keyword
			if is_key /= Void then
				is_key.process (Current)
				is_key ?= modified_node
				check modified_node /= Void implies is_key /= Void end
			end

			index := a_feature.first_indexing
			if index /= Void then
				index.process (Current)
				index ?= modified_node
				check modified_node /= Void implies index /= Void end
			end

			obsolete_message := a_feature.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			preconditions := a_feature.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			a_feature.deferred_keyword.process (Current)
			deferred_key ?= modified_node
			check deferred_key /= Void end

			postconditions := a_feature.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			a_feature.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (extended_name, args, type, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_assigner (assigner)
			res.set_is_keyword (is_key)
			res.set_first_indexing (index)
			res.set_obsolete_message (obsolete_message)
			res.set_preconditions (preconditions)
			res.set_deferred_keyword (deferred_key)
			res.set_postconditions (postconditions)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_deferred_procedure (a_feature: ET_DEFERRED_PROCEDURE) is
			-- Replace all feature components with their modified form
		local
			frozen_key, is_key, deferred_key, end_key: ET_KEYWORD
			extended_name: ET_EXTENDED_FEATURE_NAME
			args: ET_FORMAL_ARGUMENT_LIST
			index: ET_INDEXING_LIST
			obsolete_message: ET_OBSOLETE
			preconditions: ET_PRECONDITIONS
			postconditions: ET_POSTCONDITIONS
			res: ET_DEFERRED_PROCEDURE
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			args := a_feature.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			is_key := a_feature.is_keyword
			if is_key /= Void then
				is_key.process (Current)
				is_key ?= modified_node
				check modified_node /= Void implies is_key /= Void end
			end

			index := a_feature.first_indexing
			if index /= Void then
				index.process (Current)
				index ?= modified_node
				check modified_node /= Void implies index /= Void end
			end

			obsolete_message := a_feature.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			preconditions := a_feature.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			a_feature.deferred_keyword.process (Current)
			deferred_key ?= modified_node
			check deferred_key /= Void end

			postconditions := a_feature.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			a_feature.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (extended_name, args, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_is_keyword (is_key)
			res.set_first_indexing (index)
			res.set_obsolete_message (obsolete_message)
			res.set_preconditions (preconditions)
			res.set_deferred_keyword (deferred_key)
			res.set_postconditions (postconditions)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_do_function (a_feature: ET_DO_FUNCTION) is
			-- Replace all feature components with their modified form
		local
			frozen_key, is_key, end_key: ET_KEYWORD
			extended_name: ET_EXTENDED_FEATURE_NAME
			args: ET_FORMAL_ARGUMENT_LIST
			type: ET_DECLARED_TYPE
			index: ET_INDEXING_LIST
			obsolete_message: ET_OBSOLETE
			preconditions: ET_PRECONDITIONS
			locals: ET_LOCAL_VARIABLE_LIST
			compound, rescue_clause: ET_COMPOUND
			postconditions: ET_POSTCONDITIONS
			assigner: ET_ASSIGNER
			res: ET_DO_FUNCTION
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			args := a_feature.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			good_type (a_feature.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			assigner := a_feature.assigner
			if assigner /= Void then
				assigner.process (Current)
				assigner ?= modified_node
				check modified_node /= Void implies assigner /= Void end
			end

			is_key := a_feature.is_keyword
			if is_key /= Void then
				is_key.process (Current)
				is_key ?= modified_node
				check modified_node /= Void implies is_key /= Void end
			end

			index := a_feature.first_indexing
			if index /= Void then
				index.process (Current)
				index ?= modified_node
				check modified_node /= Void implies index /= Void end
			end

			obsolete_message := a_feature.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			preconditions := a_feature.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			locals := a_feature.locals
			if locals /= Void then
				locals.process (Current)
				locals ?= modified_node
				check modified_node /= Void implies locals /= Void end
			end

			compound := a_feature.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			postconditions := a_feature.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			rescue_clause := a_feature.rescue_clause
			if rescue_clause /= Void then
				rescue_clause.process (Current)
				rescue_clause ?= modified_node
				check modified_node /= Void implies rescue_clause /= Void end
			end

			a_feature.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (extended_name, args, type, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_assigner (assigner)
			res.set_is_keyword (is_key)
			res.set_first_indexing (index)
			res.set_obsolete_message (obsolete_message)
			res.set_preconditions (preconditions)
			res.set_locals (locals)
			res.set_compound (compound)
			res.set_postconditions (postconditions)
			res.set_rescue_clause (rescue_clause)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_do_function_inline_agent (an_expression: ET_DO_FUNCTION_INLINE_AGENT) is
			-- Replace all agent components with their modified form
		local
			agent_key, end_key: ET_KEYWORD
			formals: ET_FORMAL_ARGUMENT_LIST
			actuals: ET_AGENT_ARGUMENT_OPERAND_LIST
			type: ET_DECLARED_TYPE
			preconditions: ET_PRECONDITIONS
			locals: ET_LOCAL_VARIABLE_LIST
			postconditions: ET_POSTCONDITIONS
			compound, rescue_clause: ET_COMPOUND
			res: ET_DO_FUNCTION_INLINE_AGENT
		do
			an_expression.agent_keyword.process (Current)
			agent_key ?= modified_node
			check agent_key /= Void end

			formals := an_expression.formal_arguments
			if formals /= Void then
				formals.process (Current)
				formals ?= modified_node
				check modified_node /= Void implies formals /= Void end
			end

			good_type (an_expression.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			preconditions := an_expression.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			locals := an_expression.locals
			if locals /= Void then
				locals.process (Current)
				locals ?= modified_node
				check modified_node /= Void implies locals /= Void end
			end

			compound := an_expression.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			postconditions := an_expression.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			rescue_clause := an_expression.rescue_clause
			if rescue_clause /= Void then
				rescue_clause.process (Current)
				rescue_clause ?= modified_node
				check modified_node /= Void implies rescue_clause /= Void end
			end

			an_expression.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			actuals ?= an_expression.actual_arguments
			if actuals /= Void then
				actuals.process (Current)
				actuals ?= modified_node
				check modified_node /= Void implies actuals /= Void end
			end

			create res.make (formals, type, actuals)
			res.set_agent_keyword (agent_key)
			res.set_preconditions (preconditions)
			res.set_locals (locals)
			res.set_compound (compound)
			res.set_postconditions (postconditions)
			res.set_rescue_clause (rescue_clause)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_do_procedure (a_feature: ET_DO_PROCEDURE) is
			-- Replace all feature components with their modified form
		local
			frozen_key, is_key, end_key: ET_KEYWORD
			extended_name: ET_EXTENDED_FEATURE_NAME
			args: ET_FORMAL_ARGUMENT_LIST
			index: ET_INDEXING_LIST
			obsolete_message: ET_OBSOLETE
			preconditions: ET_PRECONDITIONS
			locals: ET_LOCAL_VARIABLE_LIST
			compound, rescue_clause: ET_COMPOUND
			postconditions: ET_POSTCONDITIONS
			res: ET_DO_PROCEDURE
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			args := a_feature.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			is_key := a_feature.is_keyword
			if is_key /= Void then
				is_key.process (Current)
				is_key ?= modified_node
				check modified_node /= Void implies is_key /= Void end
			end

			index := a_feature.first_indexing
			if index /= Void then
				index.process (Current)
				index ?= modified_node
				check modified_node /= Void implies index /= Void end
			end

			obsolete_message := a_feature.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			preconditions := a_feature.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			locals := a_feature.locals
			if locals /= Void then
				locals.process (Current)
				locals ?= modified_node
				check modified_node /= Void implies locals /= Void end
			end

			compound := a_feature.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			postconditions := a_feature.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			rescue_clause := a_feature.rescue_clause
			if rescue_clause /= Void then
				rescue_clause.process (Current)
				rescue_clause ?= modified_node
				check modified_node /= Void implies rescue_clause /= Void end
			end

			a_feature.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (extended_name, args, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_is_keyword (is_key)
			res.set_first_indexing (index)
			res.set_obsolete_message (obsolete_message)
			res.set_preconditions (preconditions)
			res.set_locals (locals)
			res.set_compound (compound)
			res.set_postconditions (postconditions)
			res.set_rescue_clause (rescue_clause)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_do_procedure_inline_agent (an_expression: ET_DO_PROCEDURE_INLINE_AGENT) is
			-- Replace all agent components with their modified form
		local
			agent_key, end_key: ET_KEYWORD
			formals: ET_FORMAL_ARGUMENT_LIST
			actuals: ET_AGENT_ARGUMENT_OPERAND_LIST
			preconditions: ET_PRECONDITIONS
			locals: ET_LOCAL_VARIABLE_LIST
			postconditions: ET_POSTCONDITIONS
			compound, rescue_clause: ET_COMPOUND
			res: ET_DO_PROCEDURE_INLINE_AGENT
		do
			an_expression.agent_keyword.process (Current)
			agent_key ?= modified_node
			check agent_key /= Void end

			formals := an_expression.formal_arguments
			if formals /= Void then
				formals.process (Current)
				formals ?= modified_node
				check modified_node /= Void implies formals /= Void end
			end

			preconditions := an_expression.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			locals := an_expression.locals
			if locals /= Void then
				locals.process (Current)
				locals ?= modified_node
				check modified_node /= Void implies locals /= Void end
			end

			compound := an_expression.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			postconditions := an_expression.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			rescue_clause := an_expression.rescue_clause
			if rescue_clause /= Void then
				rescue_clause.process (Current)
				rescue_clause ?= modified_node
				check modified_node /= Void implies rescue_clause /= Void end
			end

			an_expression.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			actuals ?= an_expression.actual_arguments
			if actuals /= Void then
				actuals.process (Current)
				actuals ?= modified_node
				check modified_node /= Void implies actuals /= Void end
			end

			create res.make (formals, actuals)
			res.set_agent_keyword (agent_key)
			res.set_preconditions (preconditions)
			res.set_locals (locals)
			res.set_compound (compound)
			res.set_postconditions (postconditions)
			res.set_rescue_clause (rescue_clause)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_dot_feature_name (a_name: ET_DOT_FEATURE_NAME) is
			-- Replace feature name with its modified form
		local
			feature_name: ET_FEATURE_NAME
		do
			a_name.feature_name.process (Current)
			feature_name ?= modified_node
			check feature_name /= Void end

			create {ET_DOT_FEATURE_NAME} modified_node.make (a_name.dot, feature_name)
		end

	process_dotnet_attribute (a_feature: ET_DOTNET_ATTRIBUTE) is
			-- Process `a_feature'.
		do
			process_attribute (a_feature)
		end

	process_dotnet_constant_attribute (a_feature: ET_DOTNET_CONSTANT_ATTRIBUTE) is
			-- Process `a_feature'.
		do
			process_constant_attribute (a_feature)
		end

	process_dotnet_function (a_feature: ET_DOTNET_FUNCTION) is
			-- Process `a_feature'.
		do
			-- TODO: support .NET
			modified_node := a_feature
		end

	process_dotnet_procedure (a_feature: ET_DOTNET_PROCEDURE) is
			-- Process `a_feature'.
		do
			-- TODO: support .NET
			modified_node := a_feature
		end

	process_elseif_part (an_elseif_part: ET_ELSEIF_PART) is
			-- Replace conditional and compound with their modified form
		local
			conditional: ET_CONDITIONAL
			compound: ET_COMPOUND
		do
			an_elseif_part.conditional.process (Current)
			conditional ?= modified_node
			check conditional /= Void end

			compound := an_elseif_part.then_compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			create {ET_ELSEIF_PART} modified_node.make (conditional, compound)
		end

	process_elseif_part_list (a_list: ET_ELSEIF_PART_LIST) is
			-- Replace each elseif part with its modified form
		local
			item: ET_ELSEIF_PART
			i, nb: INTEGER
			res: ET_ELSEIF_PART_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_equality_expression (an_expression: ET_EQUALITY_EXPRESSION) is
			-- Replace left, right and operator with their modified form
		local
			left, right: ET_EXPRESSION
			operator: ET_SYMBOL
		do
			an_expression.left.process (Current)
			left ?= modified_node
			check left /= Void end

			an_expression.operator.process (Current)
			operator ?= modified_node
			check operator /= Void end

			an_expression.right.process (Current)
			right ?= modified_node
			check right /= Void end

			create {ET_EQUALITY_EXPRESSION} modified_node.make (left, operator, right)
		end

	process_export_list (a_list: ET_EXPORT_LIST) is
			-- Replace each elseif part with its modified form
		local
			key: ET_KEYWORD
			item: ET_EXPORT
			i, nb: INTEGER
			res: ET_EXPORT_LIST
		do
			a_list.export_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_export_keyword (key)
			modified_node := res
		end

	process_expression_address (an_expression: ET_EXPRESSION_ADDRESS) is
			-- Replace expression with its modified form
		local
			expr: ET_PARENTHESIZED_EXPRESSION
		do
			an_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end
			create {ET_EXPRESSION_ADDRESS} modified_node.make (expr)
		end

	process_expression_comma (an_expression: ET_EXPRESSION_COMMA) is
			-- Replace expression with its modified form
		local
			expr: ET_EXPRESSION
		do
			an_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end
			create {ET_EXPRESSION_COMMA} modified_node.make (expr, an_expression.comma)
		end

	process_extended_feature_name_comma (a_name: ET_EXTENDED_FEATURE_NAME_COMMA) is
			-- Replace extended name with its modified form
		local
			extended_name: ET_EXTENDED_FEATURE_NAME
		do
			a_name.extended_feature_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end
			create {ET_EXTENDED_FEATURE_NAME_COMMA} modified_node.make (extended_name, a_name.comma)
		end

	process_external_function (a_feature: ET_EXTERNAL_FUNCTION) is
			-- Replace all feature components with their modified form
		local
			frozen_key, is_key, end_key: ET_KEYWORD
			extended_name: ET_EXTENDED_FEATURE_NAME
			args: ET_FORMAL_ARGUMENT_LIST
			type: ET_DECLARED_TYPE
			index: ET_INDEXING_LIST
			obsolete_message: ET_OBSOLETE
			preconditions: ET_PRECONDITIONS
			language: ET_EXTERNAL_LANGUAGE
			alias_clause: ET_EXTERNAL_ALIAS
			postconditions: ET_POSTCONDITIONS
			assigner: ET_ASSIGNER
			res: ET_EXTERNAL_FUNCTION
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			args := a_feature.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			good_type (a_feature.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			assigner := a_feature.assigner
			if assigner /= Void then
				assigner.process (Current)
				assigner ?= modified_node
				check modified_node /= Void implies assigner /= Void end
			end

			is_key := a_feature.is_keyword
			if is_key /= Void then
				is_key.process (Current)
				is_key ?= modified_node
				check modified_node /= Void implies is_key /= Void end
			end

			index := a_feature.first_indexing
			if index /= Void then
				index.process (Current)
				index ?= modified_node
				check modified_node /= Void implies index /= Void end
			end

			obsolete_message := a_feature.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			preconditions := a_feature.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			a_feature.language.process (Current)
			language ?= modified_node
			check language /= Void end

			alias_clause := a_feature.alias_clause
			if alias_clause /= Void then
				alias_clause.process (Current)
				alias_clause ?= modified_node
				check modified_node /= Void implies alias_clause /= Void end
			end

			postconditions := a_feature.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			a_feature.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (extended_name, args, type, language, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_assigner (assigner)
			res.set_is_keyword (is_key)
			res.set_first_indexing (index)
			res.set_obsolete_message (obsolete_message)
			res.set_preconditions (preconditions)
			res.set_alias_clause (alias_clause)
			res.set_postconditions (postconditions)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_external_function_inline_agent (an_expression: ET_EXTERNAL_FUNCTION_INLINE_AGENT) is
			-- Replace all agent components with their modified form
		local
			agent_key, end_key: ET_KEYWORD
			formals: ET_FORMAL_ARGUMENT_LIST
			actuals: ET_AGENT_ARGUMENT_OPERAND_LIST
			type: ET_DECLARED_TYPE
			preconditions: ET_PRECONDITIONS
			language: ET_EXTERNAL_LANGUAGE
			alias_clause: ET_EXTERNAL_ALIAS
			postconditions: ET_POSTCONDITIONS
			res: ET_EXTERNAL_FUNCTION_INLINE_AGENT
		do
			an_expression.agent_keyword.process (Current)
			agent_key ?= modified_node
			check agent_key /= Void end

			formals := an_expression.formal_arguments
			if formals /= Void then
				formals.process (Current)
				formals ?= modified_node
				check modified_node /= Void implies formals /= Void end
			end

			good_type (an_expression.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			preconditions := an_expression.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			an_expression.language.process (Current)
			language ?= modified_node
			check language /= Void end

			alias_clause := an_expression.alias_clause
			if alias_clause /= Void then
				alias_clause.process (Current)
				alias_clause ?= modified_node
				check modified_node /= Void implies alias_clause /= Void end
			end

			postconditions := an_expression.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			an_expression.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			actuals ?= an_expression.actual_arguments
			if actuals /= Void then
				actuals.process (Current)
				actuals ?= modified_node
				check modified_node /= Void implies actuals /= Void end
			end

			create res.make (formals, type, language, actuals)
			res.set_agent_keyword (agent_key)
			res.set_preconditions (preconditions)
			res.set_alias_clause (alias_clause)
			res.set_postconditions (postconditions)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_external_procedure (a_feature: ET_EXTERNAL_PROCEDURE) is
			-- Replace all feature components with their modified form
		local
			frozen_key, is_key, end_key: ET_KEYWORD
			extended_name: ET_EXTENDED_FEATURE_NAME
			args: ET_FORMAL_ARGUMENT_LIST
			index: ET_INDEXING_LIST
			obsolete_message: ET_OBSOLETE
			preconditions: ET_PRECONDITIONS
			language: ET_EXTERNAL_LANGUAGE
			alias_clause: ET_EXTERNAL_ALIAS
			postconditions: ET_POSTCONDITIONS
			res: ET_EXTERNAL_PROCEDURE
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			args := a_feature.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			is_key := a_feature.is_keyword
			if is_key /= Void then
				is_key.process (Current)
				is_key ?= modified_node
				check modified_node /= Void implies is_key /= Void end
			end

			index := a_feature.first_indexing
			if index /= Void then
				index.process (Current)
				index ?= modified_node
				check modified_node /= Void implies index /= Void end
			end

			obsolete_message := a_feature.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			preconditions := a_feature.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			a_feature.language.process (Current)
			language ?= modified_node
			check language /= Void end

			alias_clause := a_feature.alias_clause
			if alias_clause /= Void then
				alias_clause.process (Current)
				alias_clause ?= modified_node
				check modified_node /= Void implies alias_clause /= Void end
			end

			postconditions := a_feature.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			a_feature.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (extended_name, args, language, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_is_keyword (is_key)
			res.set_first_indexing (index)
			res.set_obsolete_message (obsolete_message)
			res.set_preconditions (preconditions)
			res.set_alias_clause (alias_clause)
			res.set_postconditions (postconditions)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_external_procedure_inline_agent (an_expression: ET_EXTERNAL_PROCEDURE_INLINE_AGENT) is
			-- Replace all agent components with their modified form
		local
			agent_key, end_key: ET_KEYWORD
			formals: ET_FORMAL_ARGUMENT_LIST
			actuals: ET_AGENT_ARGUMENT_OPERAND_LIST
			preconditions: ET_PRECONDITIONS
			language: ET_EXTERNAL_LANGUAGE
			alias_clause: ET_EXTERNAL_ALIAS
			postconditions: ET_POSTCONDITIONS
			res: ET_EXTERNAL_PROCEDURE_INLINE_AGENT
		do
			an_expression.agent_keyword.process (Current)
			agent_key ?= modified_node
			check agent_key /= Void end

			formals := an_expression.formal_arguments
			if formals /= Void then
				formals.process (Current)
				formals ?= modified_node
				check modified_node /= Void implies formals /= Void end
			end

			preconditions := an_expression.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			an_expression.language.process (Current)
			language ?= modified_node
			check language /= Void end

			alias_clause := an_expression.alias_clause
			if alias_clause /= Void then
				alias_clause.process (Current)
				alias_clause ?= modified_node
				check modified_node /= Void implies alias_clause /= Void end
			end

			postconditions := an_expression.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			an_expression.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			actuals ?= an_expression.actual_arguments
			if actuals /= Void then
				actuals.process (Current)
				actuals ?= modified_node
				check modified_node /= Void implies actuals /= Void end
			end

			create res.make (formals, language, actuals)
			res.set_agent_keyword (agent_key)
			res.set_preconditions (preconditions)
			res.set_alias_clause (alias_clause)
			res.set_postconditions (postconditions)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_false_constant (a_constant: ET_FALSE_CONSTANT) is
			-- Process `a_constant'.
		do
			modified_node := a_constant
		end

	process_feature_address (an_expression: ET_FEATURE_ADDRESS) is
			-- Replace feature name with its qualified form
		local
			name: ET_FEATURE_NAME
		do
			an_expression.name.process (Current)
			name ?= modified_node
			check name /= Void end
			create {ET_FEATURE_ADDRESS} modified_node.make (name)
		end

	process_feature_clause (a_feature_clause: ET_FEATURE_CLAUSE) is
			-- Replace keyword and clients with their modified form
		local
			key: ET_KEYWORD
			clients: ET_CLASS_NAME_LIST -- Should be ET_CLIENTS
			res: ET_FEATURE_CLAUSE
		do
			a_feature_clause.feature_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			clients := a_feature_clause.clients
			if clients /= Void then
				process_class_name_list (clients) -- clients.process (Current)
				clients ?= modified_node
				check modified_node /= Void clients /= Void end
			end

			create res.make (clients)
			res.set_feature_keyword (key)
			modified_node := res
		end

	process_feature_clause_list (a_list: ET_FEATURE_CLAUSE_LIST) is
			-- Replace each feature clause with its modified form
		local
			item: ET_FEATURE_CLAUSE
			i, nb: INTEGER
			res: ET_FEATURE_CLAUSE_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_feature_export (an_export: ET_FEATURE_EXPORT) is
			-- Replace clients and each feature name with its modified form
		local
			clients: ET_CLIENTS
			item: ET_FEATURE_NAME_ITEM
			i, nb: INTEGER
			res: ET_FEATURE_EXPORT
		do
			an_export.clients_clause.process (Current)
			clients ?= modified_node
			check clients /= Void end

			nb := an_export.count
			create res.make_with_capacity (clients, nb)
			from i := nb until i < 1 loop
				an_export.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_feature_name_comma (a_name: ET_FEATURE_NAME_COMMA) is
			-- Replace feature name with its modified form
		local
			name: ET_FEATURE_NAME
		do
			a_name.feature_name.process (Current)
			name ?= modified_node
			check name /= Void end
			create {ET_FEATURE_NAME_COMMA} modified_node.make (name, a_name.comma)
		end

	process_formal_argument (an_argument: ET_FORMAL_ARGUMENT) is
			-- Replace name and type with their modified form
		local
			name: ET_ARGUMENT_NAME
			type: ET_DECLARED_TYPE
		do
			an_argument.name_item.process (Current)
			name ?= modified_node
			check name /= Void end

			good_type (an_argument.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			create {ET_FORMAL_ARGUMENT} modified_node.make (name, type)
		end

	process_formal_argument_semicolon (an_argument: ET_FORMAL_ARGUMENT_SEMICOLON) is
			-- Replace formal argument with its modified form
		local
			formal: ET_FORMAL_ARGUMENT
		do
			an_argument.formal_argument.process (Current)
			formal ?= modified_node
			check formal /= Void end

			create {ET_FORMAL_ARGUMENT_SEMICOLON} modified_node.make (formal, an_argument.semicolon)
		end

	process_formal_argument_list (a_list: ET_FORMAL_ARGUMENT_LIST) is
			-- Replace each argument with its modified form
		local
			item: ET_FORMAL_ARGUMENT_ITEM
			i, nb: INTEGER
			res: ET_FORMAL_ARGUMENT_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_formal_comma_argument (an_argument: ET_FORMAL_COMMA_ARGUMENT) is
			-- Process `an_argument'.
		local
			arg: ET_FORMAL_ARGUMENT
		do
			process_formal_argument (an_argument)
			arg ?= modified_node
			check arg /= Void end

			create {ET_FORMAL_COMMA_ARGUMENT} modified_node.make (arg.name_item, arg.type)
		end

	process_formal_parameter (a_parameter: ET_FORMAL_PARAMETER) is
			-- Replace name and type mark with their modified form
		local
			type_mark: ET_KEYWORD
			name: ET_IDENTIFIER
			res: ET_FORMAL_PARAMETER
		do
			type_mark := a_parameter.type_mark
			if type_mark /= Void then
				type_mark.process (Current)
				type_mark ?= modified_node
				check modified_node /= Void implies type_mark /= Void end
			end

			a_parameter.name.process (Current)
			name ?= modified_node
			check name /= Void end

			create res.make (name)
			res.set_type_mark (type_mark)
			modified_node := res
		end

	process_formal_parameter_comma (a_parameter: ET_FORMAL_PARAMETER_COMMA) is
			-- Replace formal parameter with its modified form
		local
			param: ET_FORMAL_PARAMETER
		do
			a_parameter.formal_parameter.process (Current)
			param ?= modified_node
			check param /= Void end

			create {ET_FORMAL_PARAMETER_COMMA} modified_node.make (param, a_parameter.comma)
		end

	process_formal_parameter_list (a_list: ET_FORMAL_PARAMETER_LIST) is
			-- Replace each parameter with its modified form
		local
			item: ET_FORMAL_PARAMETER_ITEM
			i, nb: INTEGER
			res: ET_FORMAL_PARAMETER_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_formal_parameter_type (a_type: ET_FORMAL_PARAMETER_TYPE) is
			-- Replace name with its modified form
		local
			name: ET_IDENTIFIER
		do
			a_type.name.process (Current)
			name ?= modified_node
			check name /= Void end
			create {ET_FORMAL_PARAMETER_TYPE} modified_node.make (name, a_type.index)
		end

	process_free_operator (an_operator: ET_FREE_OPERATOR) is
			-- Process `an_operator'.
		do
			modified_node := an_operator
		end

	process_generic_class_type (a_type: ET_GENERIC_CLASS_TYPE) is
			-- Replace type mark, name and actual parameters with their modified form
		local
			type_mark: ET_KEYWORD
			name: ET_CLASS_NAME
			params: ET_ACTUAL_PARAMETER_LIST
		do
			type_mark := a_type.type_mark
			if type_mark /= Void then
				type_mark.process (Current)
				type_mark ?= modified_node
				check modified_node /= Void implies type_mark /= Void end
			end

			a_type.name.process (Current)
			name ?= modified_node
			check name /= Void end

			a_type.actual_parameters.process (Current)
			params ?= modified_node
			check params /= Void end

			create {ET_GENERIC_CLASS_TYPE} modified_node.make (type_mark, name, params, universe.eiffel_class (name))
		end

	process_hexadecimal_integer_constant (a_constant: ET_HEXADECIMAL_INTEGER_CONSTANT) is
			-- Replace cast type and sign with their modified form
		local
			type: ET_TARGET_TYPE
			sign: ET_SYMBOL_OPERATOR
			res: ET_HEXADECIMAL_INTEGER_CONSTANT
		do
			type ?= good_type (a_constant.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'

			sign := a_constant.sign
			if sign /= Void then
				sign.process (Current)
				sign ?= modified_node
				check modified_node /= Void implies sign /= Void end
			end

			create res.make (a_constant.literal)
			res.set_cast_type (type)
			res.set_sign (sign)
			modified_node := res
		end

	process_identifier (an_identifier: ET_IDENTIFIER) is
			-- Process `an_identifier'.
		do
			modified_node := an_identifier
		end

	process_identifier_colon (an_identifier: ET_IDENTIFIER_COLON) is
			-- Replace identifier with its modified form
		local
			id: ET_IDENTIFIER
		do
			an_identifier.identifier.process (Current)
			id ?= modified_node
			check id /= Void end

			create {ET_IDENTIFIER_COLON} modified_node.make (id, an_identifier.colon)
		end

	process_identifier_comma (an_identifier: ET_IDENTIFIER_COMMA) is
			-- Replace identifier with its modified form
		local
			id: ET_IDENTIFIER
		do
			an_identifier.identifier.process (Current)
			id ?= modified_node
			check id /= Void end

			create {ET_IDENTIFIER_COMMA} modified_node.make (id, an_identifier.comma)
		end

	process_if_instruction (an_instruction: ET_IF_INSTRUCTION) is
			-- Process `an_instruction'.
		local
			conditional: ET_CONDITIONAL
			elseif_parts: ET_ELSEIF_PART_LIST
			then_compound, else_compound: ET_COMPOUND
			end_key: ET_KEYWORD
			res: ET_IF_INSTRUCTION
		do
			an_instruction.conditional.process (Current)
			conditional ?= modified_node
			check conditional /= Void end

			then_compound := an_instruction.then_compound
			if then_compound /= Void then
				then_compound.process (Current)
				then_compound ?= modified_node
				check modified_node /= Void implies then_compound /= Void end
			end

			elseif_parts := an_instruction.elseif_parts
			if elseif_parts /= Void then
				elseif_parts.process (Current)
				elseif_parts ?= modified_node
				check modified_node /= Void implies elseif_parts /= Void end
			end

			else_compound := an_instruction.else_compound
			if else_compound /= Void then
				else_compound.process (Current)
				else_compound ?= modified_node
				check modified_node /= Void implies else_compound /= Void end
			end

			an_instruction.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (conditional, then_compound)
			res.set_elseif_parts (elseif_parts)
			res.set_else_compound (else_compound)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_indexing (an_indexing: ET_INDEXING) is
			-- Replace terms with their modified form
		local
			terms: ET_INDEXING_TERM_LIST
		do
			an_indexing.terms.process (Current)
			terms ?= modified_node
			check terms /= Void end

			create {ET_INDEXING} modified_node.make (terms)
		end

	process_indexing_list (a_list: ET_INDEXING_LIST) is
			-- Replace keyword and each indexing item with its modified form
		local
			key: ET_KEYWORD
			item: ET_INDEXING_ITEM
			i, nb: INTEGER
			res: ET_INDEXING_LIST
		do
			a_list.indexing_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_indexing_keyword (key)
			modified_node := res
		end

	process_indexing_semicolon (an_indexing: ET_INDEXING_SEMICOLON) is
			-- Replace indexing with its modified form
		local
			index: ET_INDEXING_ITEM
		do
			an_indexing.indexing_item.process (Current)
			index ?= modified_node
			check index /= Void end

			create {ET_INDEXING_SEMICOLON} modified_node.make (index, an_indexing.semicolon)
		end

	process_indexing_term_comma (an_indexing_term: ET_INDEXING_TERM_COMMA) is
			-- Replace indexing term with its modified form
		local
			term: ET_INDEXING_TERM
		do
			an_indexing_term.indexing_term.process (Current)
			term ?= modified_node
			check term /= Void end

			create {ET_INDEXING_TERM_COMMA} modified_node.make (term, an_indexing_term.comma)
		end

	process_indexing_term_list (a_list: ET_INDEXING_TERM_LIST) is
			-- Replace each indexing term with its modified form
		local
			item: ET_INDEXING_TERM_ITEM
			i, nb: INTEGER
			res: ET_INDEXING_TERM_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_infix_cast_expression (an_expression: ET_INFIX_CAST_EXPRESSION) is
			-- Replace expression and type with their modified form
		local
			expr: ET_EXPRESSION
			type: ET_TYPE
		do
			an_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end

			good_type (an_expression.type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			create {ET_INFIX_CAST_EXPRESSION} modified_node.make (expr, type)
		end

	process_infix_and_then_operator (an_operator: ET_INFIX_AND_THEN_OPERATOR) is
			-- Replace "and" and "then" keywords with their modified form
		local
			and_key, then_key: ET_KEYWORD
			res: ET_INFIX_AND_THEN_OPERATOR
		do
			an_operator.and_keyword.process (Current)
			and_key ?= modified_node
			check and_key /= Void end

			an_operator.then_keyword.process (Current)
			then_key ?= modified_node
			check then_key /= Void end

			create res.make
			res.set_and_keyword (and_key)
			res.set_then_keyword (then_key)
			modified_node := res
		end

	process_infix_expression (an_expression: ET_INFIX_EXPRESSION) is
			-- Replace left, name and right with their modified form
		local
			left, right: ET_EXPRESSION
			name: ET_OPERATOR
		do
			an_expression.left.process (Current)
			left ?= modified_node
			check left /= Void end

			an_expression.name.process (Current)
			name ?= modified_node
			check name /= Void end

			an_expression.right.process (Current)
			right ?= modified_node
			check right /= Void end

			create {ET_INFIX_EXPRESSION} modified_node.make (left, name, right)
		end

	process_infix_free_name (a_name: ET_INFIX_FREE_NAME) is
			-- Process `a_name'.
		do
			process_infix_name (a_name)
		end

	process_infix_name (a_name: ET_INFIX_NAME) is
			-- Replace keyword and name with their modified form
		do
			process_alias_name (a_name)
		end

	process_infix_or_else_operator (an_operator: ET_INFIX_OR_ELSE_OPERATOR) is
			-- Replace "or" and "else" keywords with their modified form
		local
			or_key, else_key: ET_KEYWORD
			res: ET_INFIX_OR_ELSE_OPERATOR
		do
			an_operator.or_keyword.process (Current)
			or_key ?= modified_node
			check or_key /= Void end

			an_operator.else_keyword.process (Current)
			else_key ?= modified_node
			check else_key /= Void end

			create res.make
			res.set_or_keyword (or_key)
			res.set_else_keyword (else_key)
			modified_node := res
		end

	process_inspect_instruction (an_instruction: ET_INSPECT_INSTRUCTION) is
			-- Replace conditional, when parts, else compound and end keyword with their modified form
		local
			conditional: ET_CONDITIONAL
			when_parts: ET_WHEN_PART_LIST
			else_compound: ET_COMPOUND
			end_key: ET_KEYWORD
			res: ET_INSPECT_INSTRUCTION
		do
			an_instruction.conditional.process (Current)
			conditional ?= modified_node
			check conditional /= Void end

			when_parts := an_instruction.when_parts
			if when_parts /= Void then
				when_parts.process (Current)
				when_parts ?= modified_node
				check modified_node /= Void implies when_parts /= Void end
			end

			else_compound := an_instruction.else_compound
			if else_compound /= Void then
				else_compound.process (Current)
				else_compound ?= modified_node
				check modified_node /= Void implies else_compound /= Void end
			end

			an_instruction.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (conditional, when_parts)
			res.set_else_compound (else_compound)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_invariants (a_list: ET_INVARIANTS) is
			-- Replace keyword and each invariant with its modified form
		local
			key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_ASSERTION_ITEM
			res: ET_INVARIANTS
		do
			a_list.invariant_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (a_list.implementation_class, nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_invariant_keyword (key)
			modified_node := res
		end

	process_keyword (a_keyword: ET_KEYWORD) is
			-- Process `a_keyword'.
		do
			modified_node := a_keyword
		end

	process_keyword_expression (an_expression: ET_KEYWORD_EXPRESSION) is
			-- Replace keyword and expression with their modified form
		local
			key: ET_KEYWORD
			expr: ET_EXPRESSION
		do
			an_expression.keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			an_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end

			create {ET_KEYWORD_EXPRESSION} modified_node.make (key, expr)
		end

	process_keyword_feature_name_list (a_list: ET_KEYWORD_FEATURE_NAME_LIST) is
			-- Replace keyword and each name with its modified form
		local
			key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_FEATURE_NAME_ITEM
			res: ET_KEYWORD_FEATURE_NAME_LIST
		do
			a_list.keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_keyword (key)
			modified_node := res
		end

	process_keyword_manifest_string (a_string: ET_KEYWORD_MANIFEST_STRING) is
			-- Replace keyword and string with their modified form
		local
			key: ET_KEYWORD
			str: ET_MANIFEST_STRING
		do
			a_string.keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			a_string.manifest_string.process (Current)
			str ?= modified_node
			check str /= Void end

			create {ET_KEYWORD_MANIFEST_STRING} modified_node.make (key, str)
		end

	process_keyword_operator (a_keyword: ET_KEYWORD_OPERATOR) is
			-- Process `a_keyword'.
		do
			modified_node := a_keyword
		end

	process_labeled_actual_parameter (a_parameter: ET_LABELED_ACTUAL_PARAMETER) is
			-- Replace label and type with their modified form
		local
			label: ET_LABEL
			type: ET_DECLARED_TYPE
		do
			a_parameter.label_item.process (Current)
			label ?= modified_node
			check label /= Void end

			good_type (a_parameter.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			create {ET_LABELED_ACTUAL_PARAMETER} modified_node.make (label, type)
		end

	process_labeled_actual_parameter_semicolon (a_parameter: ET_LABELED_ACTUAL_PARAMETER_SEMICOLON) is
			-- Replace parameter with its modified form
		local
			param: ET_LABELED_ACTUAL_PARAMETER
		do
			a_parameter.actual_parameter.process (Current)
			param ?= modified_node
			check param /= Void end

			create {ET_LABELED_ACTUAL_PARAMETER_SEMICOLON} modified_node.make (param, a_parameter.semicolon)
		end

	process_labeled_comma_actual_parameter (a_parameter: ET_LABELED_COMMA_ACTUAL_PARAMETER) is
			-- Process `a_parameter'.
		local
			param: ET_LABELED_ACTUAL_PARAMETER
		do
			process_labeled_actual_parameter (a_parameter)
			param ?= modified_node
			check param /= Void end

			create {ET_LABELED_COMMA_ACTUAL_PARAMETER} modified_node.make (param.label_item, param.declared_type)
		end

	process_like_current (a_type: ET_LIKE_CURRENT) is
			-- Replace keywords with their modified form
		local
			like_key: ET_KEYWORD
			cur: ET_CURRENT
			res: ET_LIKE_CURRENT
		do
			a_type.like_keyword.process (Current)
			like_key ?= modified_node
			check like_key /= Void end

			a_type.current_keyword.process (Current)
			cur ?= modified_node
			check cur /= Void end

			create res.make
			res.set_like_keyword (like_key)
			res.set_current_keyword (cur)
			modified_node := res
		end

	process_like_feature (a_type: ET_LIKE_FEATURE) is
			-- Replace keyword and feature name with their modified form
		local
			like_key: ET_KEYWORD
			name: ET_FEATURE_NAME
			res: ET_LIKE_FEATURE
		do
			a_type.like_keyword.process (Current)
			like_key ?= modified_node
			check like_key /= Void end

			a_type.name.process (Current)
			name ?= modified_node
			check name /= Void end

			create res.make_with_seed (name, a_type.seed)
			res.set_like_keyword (like_key)
			modified_node := res
		end

	process_local_variable_semicolon (a_local: ET_LOCAL_VARIABLE_SEMICOLON) is
			-- Replace variable with its modified form
		local
			var: ET_LOCAL_VARIABLE
		do
			a_local.local_variable.process (Current)
			var ?= modified_node
			check var /= Void end

			create {ET_LOCAL_VARIABLE_SEMICOLON} modified_node.make (var, a_local.semicolon)
		end

	process_local_comma_variable (a_local: ET_LOCAL_COMMA_VARIABLE) is
			-- Process `a_local'.
		local
			var: ET_LOCAL_VARIABLE
		do
			process_local_variable (a_local)
			var ?= modified_node
			check var /= Void end

			create {ET_LOCAL_COMMA_VARIABLE} modified_node.make (var.name_item, var.declared_type)
		end

	process_local_variable (a_local: ET_LOCAL_VARIABLE) is
			-- Replace name and type with their modified form
		local
			name: ET_LOCAL_NAME
			type: ET_DECLARED_TYPE
		do
			a_local.name_item.process (Current)
			name ?= modified_node
			check name /= Void end

			good_type (a_local.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			create {ET_LOCAL_VARIABLE} modified_node.make (name, type)
		end

	process_local_variable_list (a_list: ET_LOCAL_VARIABLE_LIST) is
			-- Replace keyword and each variable with its modified form
		local
			key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_LOCAL_VARIABLE_ITEM
			res: ET_LOCAL_VARIABLE_LIST
		do
			a_list.local_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_local_keyword (key)
			modified_node := res
		end

	process_loop_instruction (an_instruction: ET_LOOP_INSTRUCTION) is
			-- Replace end keyword, from, until, loop, invariant and variant parts with their modified form
		local
			invariant_part: ET_LOOP_INVARIANTS
			variant_part: ET_VARIANT
			conditional: ET_CONDITIONAL
			from_compound, loop_compound: ET_COMPOUND
			end_key: ET_KEYWORD
			res: ET_LOOP_INSTRUCTION
		do
			from_compound := an_instruction.from_compound
			if from_compound /= Void then
				from_compound.process (Current)
				from_compound ?= modified_node
				check modified_node /= Void implies from_compound /= Void end
			end

			invariant_part := an_instruction.invariant_part
			if invariant_part /= Void then
				invariant_part.process (Current)
				invariant_part ?= modified_node
				check modified_node /= Void implies invariant_part /= Void end
			end

			variant_part := an_instruction.variant_part
			if variant_part /= Void then
				variant_part.process (Current)
				variant_part ?= modified_node
				check modified_node /= Void implies variant_part /= Void end
			end

			an_instruction.until_conditional.process (Current)
			conditional ?= modified_node
			check conditional /= Void end

			loop_compound := an_instruction.loop_compound
			if loop_compound /= Void then
				loop_compound.process (Current)
				loop_compound ?= modified_node
				check modified_node /= Void implies loop_compound /= Void end
			end

			an_instruction.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (from_compound, conditional, loop_compound)
			res.set_invariant_part (invariant_part)
			res.set_variant_part (variant_part)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_loop_invariants (a_list: ET_LOOP_INVARIANTS) is
			-- Replace keyword and each invariant with its modified form
		local
			key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_ASSERTION_ITEM
			res: ET_LOOP_INVARIANTS
		do
			a_list.invariant_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_invariant_keyword (key)
			modified_node := res
		end

	process_manifest_array (an_expression: ET_MANIFEST_ARRAY) is
			-- Replace each expression with its modified form
		local
			i, nb: INTEGER
			item: ET_EXPRESSION_ITEM
			res: ET_MANIFEST_ARRAY
		do
			nb := an_expression.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				an_expression.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_manifest_string_comma (a_string: ET_MANIFEST_STRING_COMMA) is
			-- Replace string with its modified form
		local
			str: ET_MANIFEST_STRING
		do
			a_string.manifest_string.process (Current)
			str ?= modified_node
			check str /= Void end

			create {ET_MANIFEST_STRING_COMMA} modified_node.make (str, a_string.comma)
		end

	process_manifest_string_list (a_list: ET_MANIFEST_STRING_LIST) is
			-- Replace each string with its modified form
		local
			i, nb: INTEGER
			item: ET_MANIFEST_STRING_ITEM
			res: ET_MANIFEST_STRING_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_manifest_tuple (an_expression: ET_MANIFEST_TUPLE) is
			-- Replace each expression with its modified form
		local
			i, nb: INTEGER
			item: ET_EXPRESSION_ITEM
			res: ET_MANIFEST_ARRAY
		do
			nb := an_expression.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				an_expression.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

	process_manifest_type (an_expression: ET_MANIFEST_TYPE) is
			-- Replace type eith its modified form
		local
			type: ET_TYPE
		do
			good_type (an_expression.type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			create {ET_MANIFEST_TYPE} modified_node.make (type)
		end

	process_old_expression (an_expression: ET_OLD_EXPRESSION) is
			-- Replace keyword and expression with their modified form
		local
			old_key: ET_KEYWORD
			expr: ET_EXPRESSION
			res: ET_OLD_EXPRESSION
		do
			an_expression.old_keyword.process (Current)
			old_key ?= modified_node
			check old_key /= Void end

			an_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end

			create res.make (expr)
			res.set_old_keyword (old_key)
			modified_node := res
		end

	process_once_function (a_feature: ET_ONCE_FUNCTION) is
			-- Replace all feature components with their modified form
		local
			frozen_key, is_key, end_key: ET_KEYWORD
			extended_name: ET_EXTENDED_FEATURE_NAME
			args: ET_FORMAL_ARGUMENT_LIST
			type: ET_DECLARED_TYPE
			index: ET_INDEXING_LIST
			obsolete_message: ET_OBSOLETE
			preconditions: ET_PRECONDITIONS
			locals: ET_LOCAL_VARIABLE_LIST
			compound, rescue_clause: ET_COMPOUND
			postconditions: ET_POSTCONDITIONS
			assigner: ET_ASSIGNER
			res: ET_ONCE_FUNCTION
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			args := a_feature.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			good_type (a_feature.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			assigner := a_feature.assigner
			if assigner /= Void then
				assigner.process (Current)
				assigner ?= modified_node
				check modified_node /= Void implies assigner /= Void end
			end

			is_key := a_feature.is_keyword
			if is_key /= Void then
				is_key.process (Current)
				is_key ?= modified_node
				check modified_node /= Void implies is_key /= Void end
			end

			index := a_feature.first_indexing
			if index /= Void then
				index.process (Current)
				index ?= modified_node
				check modified_node /= Void implies index /= Void end
			end

			obsolete_message := a_feature.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			preconditions := a_feature.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			locals := a_feature.locals
			if locals /= Void then
				locals.process (Current)
				locals ?= modified_node
				check modified_node /= Void implies locals /= Void end
			end

			compound := a_feature.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			postconditions := a_feature.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			rescue_clause := a_feature.rescue_clause
			if rescue_clause /= Void then
				rescue_clause.process (Current)
				rescue_clause ?= modified_node
				check modified_node /= Void implies rescue_clause /= Void end
			end

			a_feature.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (extended_name, args, type, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_assigner (assigner)
			res.set_is_keyword (is_key)
			res.set_first_indexing (index)
			res.set_obsolete_message (obsolete_message)
			res.set_preconditions (preconditions)
			res.set_locals (locals)
			res.set_compound (compound)
			res.set_postconditions (postconditions)
			res.set_rescue_clause (rescue_clause)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_once_function_inline_agent (an_expression: ET_ONCE_FUNCTION_INLINE_AGENT) is
			-- Replace all agent components with their modified form
		local
			agent_key, end_key: ET_KEYWORD
			formals: ET_FORMAL_ARGUMENT_LIST
			actuals: ET_AGENT_ARGUMENT_OPERAND_LIST
			type: ET_DECLARED_TYPE
			preconditions: ET_PRECONDITIONS
			locals: ET_LOCAL_VARIABLE_LIST
			postconditions: ET_POSTCONDITIONS
			compound, rescue_clause: ET_COMPOUND
			res: ET_ONCE_FUNCTION_INLINE_AGENT
		do
			an_expression.agent_keyword.process (Current)
			agent_key ?= modified_node
			check agent_key /= Void end

			formals := an_expression.formal_arguments
			if formals /= Void then
				formals.process (Current)
				formals ?= modified_node
				check modified_node /= Void implies formals /= Void end
			end

			good_type (an_expression.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			preconditions := an_expression.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			locals := an_expression.locals
			if locals /= Void then
				locals.process (Current)
				locals ?= modified_node
				check modified_node /= Void implies locals /= Void end
			end

			compound := an_expression.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			postconditions := an_expression.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			rescue_clause := an_expression.rescue_clause
			if rescue_clause /= Void then
				rescue_clause.process (Current)
				rescue_clause ?= modified_node
				check modified_node /= Void implies rescue_clause /= Void end
			end

			an_expression.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			actuals ?= an_expression.actual_arguments
			if actuals /= Void then
				actuals.process (Current)
				actuals ?= modified_node
				check modified_node /= Void implies actuals /= Void end
			end

			create res.make (formals, type, actuals)
			res.set_agent_keyword (agent_key)
			res.set_preconditions (preconditions)
			res.set_locals (locals)
			res.set_compound (compound)
			res.set_postconditions (postconditions)
			res.set_rescue_clause (rescue_clause)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_once_manifest_string (an_expression: ET_ONCE_MANIFEST_STRING) is
			-- Replace keyword and string with their modified form
		local
			key: ET_KEYWORD
			str: ET_MANIFEST_STRING
			res: ET_ONCE_MANIFEST_STRING
		do
			an_expression.once_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			an_expression.manifest_string.process (Current)
			str ?= modified_node
			check str /= Void end

			create res.make (str)
			res.set_once_keyword (key)
			modified_node := res
		end

	process_once_procedure (a_feature: ET_ONCE_PROCEDURE) is
			-- Replace all feature components with their modified form
		local
			frozen_key, is_key, end_key: ET_KEYWORD
			extended_name: ET_EXTENDED_FEATURE_NAME
			args: ET_FORMAL_ARGUMENT_LIST
			index: ET_INDEXING_LIST
			obsolete_message: ET_OBSOLETE
			preconditions: ET_PRECONDITIONS
			locals: ET_LOCAL_VARIABLE_LIST
			compound, rescue_clause: ET_COMPOUND
			postconditions: ET_POSTCONDITIONS
			res: ET_ONCE_PROCEDURE
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end

			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			args := a_feature.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			is_key := a_feature.is_keyword
			if is_key /= Void then
				is_key.process (Current)
				is_key ?= modified_node
				check modified_node /= Void implies is_key /= Void end
			end

			index := a_feature.first_indexing
			if index /= Void then
				index.process (Current)
				index ?= modified_node
				check modified_node /= Void implies index /= Void end
			end

			obsolete_message := a_feature.obsolete_message
			if obsolete_message /= Void then
				obsolete_message.process (Current)
				obsolete_message ?= modified_node
				check modified_node /= Void implies obsolete_message /= Void end
			end

			preconditions := a_feature.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			locals := a_feature.locals
			if locals /= Void then
				locals.process (Current)
				locals ?= modified_node
				check modified_node /= Void implies locals /= Void end
			end

			compound := a_feature.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			postconditions := a_feature.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			rescue_clause := a_feature.rescue_clause
			if rescue_clause /= Void then
				rescue_clause.process (Current)
				rescue_clause ?= modified_node
				check modified_node /= Void implies rescue_clause /= Void end
			end

			a_feature.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			create res.make (extended_name, args, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_is_keyword (is_key)
			res.set_first_indexing (index)
			res.set_obsolete_message (obsolete_message)
			res.set_preconditions (preconditions)
			res.set_locals (locals)
			res.set_compound (compound)
			res.set_postconditions (postconditions)
			res.set_rescue_clause (rescue_clause)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_once_procedure_inline_agent (an_expression: ET_ONCE_PROCEDURE_INLINE_AGENT) is
			-- Replace all agent components with their modified form
		local
			agent_key, end_key: ET_KEYWORD
			formals: ET_FORMAL_ARGUMENT_LIST
			actuals: ET_AGENT_ARGUMENT_OPERAND_LIST
			preconditions: ET_PRECONDITIONS
			locals: ET_LOCAL_VARIABLE_LIST
			postconditions: ET_POSTCONDITIONS
			compound, rescue_clause: ET_COMPOUND
			res: ET_ONCE_PROCEDURE_INLINE_AGENT
		do
			an_expression.agent_keyword.process (Current)
			agent_key ?= modified_node
			check agent_key /= Void end

			formals := an_expression.formal_arguments
			if formals /= Void then
				formals.process (Current)
				formals ?= modified_node
				check modified_node /= Void implies formals /= Void end
			end

			preconditions := an_expression.preconditions
			if preconditions /= Void then
				preconditions.process (Current)
				preconditions ?= modified_node
				check modified_node /= Void implies preconditions /= Void end
			end

			locals := an_expression.locals
			if locals /= Void then
				locals.process (Current)
				locals ?= modified_node
				check modified_node /= Void implies locals /= Void end
			end

			compound := an_expression.compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			postconditions := an_expression.postconditions
			if postconditions /= Void then
				postconditions.process (Current)
				postconditions ?= modified_node
				check modified_node /= Void implies postconditions /= Void end
			end

			rescue_clause := an_expression.rescue_clause
			if rescue_clause /= Void then
				rescue_clause.process (Current)
				rescue_clause ?= modified_node
				check modified_node /= Void implies rescue_clause /= Void end
			end

			an_expression.end_keyword.process (Current)
			end_key ?= modified_node
			check end_key /= Void end

			actuals ?= an_expression.actual_arguments
			if actuals /= Void then
				actuals.process (Current)
				actuals ?= modified_node
				check modified_node /= Void implies actuals /= Void end
			end

			create res.make (formals, actuals)
			res.set_agent_keyword (agent_key)
			res.set_preconditions (preconditions)
			res.set_locals (locals)
			res.set_compound (compound)
			res.set_postconditions (postconditions)
			res.set_rescue_clause (rescue_clause)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_parent (a_parent: ET_PARENT) is
			-- Replace all parent components with their modified form
		local
			type: ET_BASE_TYPE
			renames: ET_RENAME_LIST
			exports: ET_EXPORT_LIST
			undefines: ET_KEYWORD_FEATURE_NAME_LIST
			redefines: ET_KEYWORD_FEATURE_NAME_LIST
			selects: ET_KEYWORD_FEATURE_NAME_LIST
			end_key: ET_KEYWORD
			res: ET_PARENT
		do
			a_parent.type.process (Current)
			type ?= modified_node
			check type /= Void end

			renames := a_parent.renames
			if renames /= Void then
				renames.process (Current)
				renames ?= modified_node
				check modified_node /= Void implies renames /= Void end
			end

			exports := a_parent.exports
			if exports /= Void then
				exports.process (Current)
				exports ?= modified_node
				check modified_node /= Void implies exports /= Void end
			end

			undefines := a_parent.undefines
			if undefines /= Void then
				undefines.process (Current)
				undefines ?= modified_node
				check modified_node /= Void implies undefines /= Void end
			end

			redefines := a_parent.redefines
			if redefines /= Void then
				redefines.process (Current)
				redefines ?= modified_node
				check modified_node /= Void implies redefines /= Void end
			end

			selects := a_parent.selects
			if selects /= Void then
				selects.process (Current)
				selects ?= modified_node
				check modified_node /= Void implies selects /= Void end
			end

			end_key := a_parent.end_keyword
			if end_key /= Void then
				end_key.process (Current)
				end_key ?= modified_node
				check modified_node /= Void implies end_key /= Void end
			end

			create res.make (type, renames, exports, undefines, redefines, selects)
			res.set_end_keyword (end_key)
			modified_node := res
		end

	process_parent_semicolon (a_parent: ET_PARENT_SEMICOLON) is
			-- Replace parent with its modified form
		local
			parent: ET_PARENT
		do
			a_parent.parent.process (Current)
			parent ?= modified_node
			check parent /= Void end

			create {ET_PARENT_SEMICOLON} modified_node.make (parent, a_parent.semicolon)
		end

	process_parenthesized_expression (an_expression: ET_PARENTHESIZED_EXPRESSION) is
			-- Replace nested expression with its modified form
		local
			expr: ET_EXPRESSION
		do
			an_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end
			create {ET_PARENTHESIZED_EXPRESSION} modified_node.make (expr)
		end

	process_parent_list (a_list: ET_PARENT_LIST) is
			-- Replace keyword and each parent with its modified form
		local
			key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_PARENT_ITEM
			res: ET_PARENT_LIST
		do
			a_list.inherit_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_inherit_keyword (key)
			modified_node := res
		end

	process_postconditions (a_list: ET_POSTCONDITIONS) is
			-- Replace keywords and each postcondition with its modified form
		local
			ensure_key, then_key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_ASSERTION_ITEM
			res: ET_POSTCONDITIONS
		do
			a_list.ensure_keyword.process (Current)
			ensure_key ?= modified_node
			check ensure_key /= Void end

			then_key := a_list.then_keyword
			if then_key /= Void then
				then_key.process (Current)
				then_key ?= modified_node
				check modified_node /= Void implies then_key /= Void end
			end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_ensure_keyword (ensure_key)
			res.set_then_keyword (then_key)
			modified_node := res
		end

	process_preconditions (a_list: ET_PRECONDITIONS) is
			-- Replace keywords and each precondition with its modified form
		local
			require_key, else_key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_ASSERTION_ITEM
			res: ET_PRECONDITIONS
		do
			a_list.require_keyword.process (Current)
			require_key ?= modified_node
			check require_key /= Void end

			else_key := a_list.else_keyword
			if else_key /= Void then
				else_key.process (Current)
				else_key ?= modified_node
				check modified_node /= Void implies else_key /= Void end
			end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_require_keyword (require_key)
			res.set_else_keyword (else_key)
			modified_node := res
		end

	process_precursor_expression (an_expression: ET_PRECURSOR_EXPRESSION) is
			-- Replace keyword, parent name and arguments with their modified form
		local
			key: ET_PRECURSOR_KEYWORD
			parent_name: ET_PRECURSOR_CLASS_NAME
			args: ET_ACTUAL_ARGUMENT_LIST
			res: ET_PRECURSOR_EXPRESSION
		do
			an_expression.precursor_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			parent_name := an_expression.parent_name
			if parent_name /= Void then
				parent_name.process (Current)
				parent_name ?= modified_node
				check modified_node /= Void implies parent_name /= Void end
			end

			args := an_expression.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			create res.make (parent_name, args)
			res.set_precursor_keyword (key)
			modified_node := res
		end

	process_precursor_instruction (an_instruction: ET_PRECURSOR_INSTRUCTION) is
			-- Replace keyword, parent name and arguments with their mosidied form
		local
			key: ET_PRECURSOR_KEYWORD
			parent_name: ET_PRECURSOR_CLASS_NAME
			args: ET_ACTUAL_ARGUMENT_LIST
			res: ET_PRECURSOR_INSTRUCTION
		do
			an_instruction.precursor_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			parent_name := an_instruction.parent_name
			if parent_name /= Void then
				parent_name.process (Current)
				parent_name ?= modified_node
				check modified_node /= Void implies parent_name /= Void end
			end

			args := an_instruction.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			create res.make (parent_name, args)
			res.set_precursor_keyword (key)
			modified_node := res
		end

	process_precursor_keyword (a_keyword: ET_PRECURSOR_KEYWORD) is
			-- Process `a_keyword'.
		do
			modified_node := a_keyword
		end

	process_prefix_expression (an_expression: ET_PREFIX_EXPRESSION) is
			-- Replace operator name and expression with their modified
		local
			name: ET_OPERATOR
			expr: ET_EXPRESSION
		do
			an_expression.name.process (Current)
			name ?= modified_node
			check name /= Void end

			an_expression.expression.process (Current)
			expr ?= modified_node
			check expr /= Void end

			create {ET_PREFIX_EXPRESSION} modified_node.make (name, expr)
		end

	process_prefix_free_name (a_name: ET_PREFIX_FREE_NAME) is
			-- Process `a_name'.
		do
			process_prefix_name (a_name)
		end

	process_prefix_name (a_name: ET_PREFIX_NAME) is
			-- Process `a_name'.
		do
			process_alias_name (a_name)
		end

	process_qualified_call (a_call: ET_QUALIFIED_CALL) is
			-- Replace qualidied name and arguments with their modified form
		local
			qualified_name: ET_QUALIFIED_FEATURE_NAME
			args: ET_ACTUAL_ARGUMENT_LIST
		do
			a_call.qualified_name.process (Current)
			qualified_name ?= modified_node
			check qualified_name /= Void end

			args := a_call.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check args /= Void end
			end

			create {ET_QUALIFIED_CALL} modified_node.make (qualified_name, args)
		end

	process_question_mark_symbol (a_symbol: ET_QUESTION_MARK_SYMBOL) is
			-- Process `a_symbol'.
		do
			modified_node := a_symbol
		end

	process_regular_integer_constant (a_constant: ET_REGULAR_INTEGER_CONSTANT) is
			-- Replace cast type and sign with their modified form
		local
			type: ET_TARGET_TYPE
			sign: ET_SYMBOL_OPERATOR
			res: ET_REGULAR_INTEGER_CONSTANT
		do
			type ?= good_type (a_constant.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'			

			sign := a_constant.sign
			if sign /= Void then
				sign.process (Current)
				sign ?= modified_node
				check modified_node /= Void implies sign /= Void end
			end

			create res.make (a_constant.literal)
			res.set_cast_type (type)
			res.set_sign (sign)
			modified_node := res
		end

	process_regular_manifest_string (a_string: ET_REGULAR_MANIFEST_STRING) is
			-- Replace cast type with its modified form
		local
			type: ET_TARGET_TYPE
			res: ET_REGULAR_MANIFEST_STRING
		do
			type ?= good_type (a_string.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'			
			create res.make (a_string.literal)
			res.set_cast_type (type)
			modified_node := res
		end

	process_regular_real_constant (a_constant: ET_REGULAR_REAL_CONSTANT) is
			-- Replace cast type and sign with their modified form
		local
			type: ET_TARGET_TYPE
			sign: ET_SYMBOL_OPERATOR
			res: ET_REGULAR_REAL_CONSTANT
		do
			type ?= good_type (a_constant.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'			

			sign := a_constant.sign
			if sign /= Void then
				sign.process (Current)
				sign ?= modified_node
				check modified_node /= Void implies sign /= Void end
			end

			create res.make (a_constant.literal)
			res.set_cast_type (type)
			res.set_sign (sign)
			modified_node := res
		end

	process_rename (a_rename: ET_RENAME) is
			-- Replace keyword and names with their modified form
		local
			old_name: ET_FEATURE_NAME
			new_name: ET_EXTENDED_FEATURE_NAME
			as_key: ET_KEYWORD
			res: ET_RENAME
		do
			a_rename.old_name.process (Current)
			old_name ?= modified_node
			check old_name /= Void end

			a_rename.as_keyword.process (Current)
			as_key ?= modified_node
			check as_key /= Void end

			a_rename.new_name.process (Current)
			new_name ?= modified_node
			check new_name /= Void end

			create res.make (old_name, new_name)
			res.set_as_keyword (as_key)
			modified_node := res
		end

	process_rename_comma (a_rename: ET_RENAME_COMMA) is
			-- Replace rename with its modified form
		local
			ren: ET_RENAME
			res: ET_RENAME_COMMA
		do
			process_rename (a_rename)
			ren ?= modified_node
			check ren /= Void end

			create res.make (ren.old_name, ren.new_name, a_rename.comma)
			res.set_as_keyword (ren.as_keyword)
			modified_node := res
		end

	process_rename_list (a_list: ET_RENAME_LIST) is
			-- Replace keyword and each rename with its modified form
		local
			key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_RENAME_ITEM
			res: ET_RENAME_LIST
		do
			a_list.rename_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_rename_keyword (key)
			modified_node := res
		end

	process_result (an_expression: ET_RESULT) is
			-- Process `an_expression'.
		do
			modified_node := an_expression
		end

	process_result_address (an_expression: ET_RESULT_ADDRESS) is
			-- Replace result with its modified form
		local
			key: ET_RESULT
			res: ET_RESULT_ADDRESS
		do
			an_expression.result_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			create res.make
			res.set_result_keyword (key)
			modified_node := res
		end

	process_retry_instruction (an_instruction: ET_RETRY_INSTRUCTION) is
			-- Process `an_instruction'.
		do
			modified_node := an_instruction
		end

	process_semicolon_symbol (a_symbol: ET_SEMICOLON_SYMBOL) is
			-- Process `a_symbol'.
		do
			modified_node := a_symbol
		end

	process_special_manifest_string (a_string: ET_SPECIAL_MANIFEST_STRING) is
			-- Replace cast type with its modified form
		local
			type: ET_TARGET_TYPE
			res: ET_SPECIAL_MANIFEST_STRING
		do
			type ?= good_type (a_string.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_string.type'			
			create res.make (a_string.literal, a_string.value)
			res.set_cast_type (type)
			modified_node := res
		end

	process_static_call_expression (an_expression: ET_STATIC_CALL_EXPRESSION) is
			-- Replace keyword, type, name and args with their modified form
		local
			feature_key: ET_KEYWORD
			args: ET_ACTUAL_ARGUMENT_LIST
			type: ET_TARGET_TYPE
			qualified_name: ET_QUALIFIED_FEATURE_NAME
			res: ET_STATIC_CALL_EXPRESSION
		do
			feature_key := an_expression.feature_keyword
			if feature_key /= Void then
				feature_key.process (Current)
				feature_key ?= modified_node
				check modified_node /= Void implies feature_key /= Void end
			end

			good_type (an_expression.static_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			an_expression.qualified_name.process (Current)
			qualified_name ?= modified_node
			check qualified_name /= Void end

			args := an_expression.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			create res.make (type, qualified_name, args)
			res.set_feature_keyword (feature_key)
			modified_node := res
		end

	process_static_call_instruction (an_instruction: ET_STATIC_CALL_INSTRUCTION) is
			-- Replace keyword, type, name and args with their modified form
		local
			feature_key: ET_KEYWORD
			args: ET_ACTUAL_ARGUMENT_LIST
			type: ET_TARGET_TYPE
			qualified_name: ET_QUALIFIED_FEATURE_NAME
			res: ET_STATIC_CALL_INSTRUCTION
		do
			feature_key := an_instruction.feature_keyword
			if feature_key /= Void then
				feature_key.process (Current)
				feature_key ?= modified_node
				check modified_node /= Void implies feature_key /= Void end
			end

			good_type (an_instruction.static_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			an_instruction.qualified_name.process (Current)
			qualified_name ?= modified_node
			check qualified_name /= Void end

			args := an_instruction.arguments
			if args /= Void then
				args.process (Current)
				args ?= modified_node
				check modified_node /= Void implies args /= Void end
			end

			create res.make (type, qualified_name, args)
			res.set_feature_keyword (feature_key)
			modified_node := res
		end

	process_strip_expression (an_expression: ET_STRIP_EXPRESSION) is
			-- Replace keyword and each name with its modified form
		local
			key: ET_KEYWORD
			i, nb: INTEGER
			item: ET_FEATURE_NAME
			res: ET_STRIP_EXPRESSION
		do
			an_expression.strip_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			nb := an_expression.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				an_expression.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			res.set_strip_keyword (key)
			modified_node := res
		end

	process_symbol (a_symbol: ET_SYMBOL) is
			-- Process `a_symbol'.
		do
			modified_node := a_symbol
		end

	process_symbol_operator (a_symbol: ET_SYMBOL_OPERATOR) is
			-- Process `a_symbol'.
		do
			modified_node := a_symbol
		end

	process_tagged_assertion (an_assertion: ET_TAGGED_ASSERTION) is
			-- Replace tag and expression with their modified form
		local
			tag: ET_TAG
			expr: ET_EXPRESSION
			res: ET_TAGGED_ASSERTION
		do
			an_assertion.tag.process (Current)
			tag ?= modified_node
			check tag /= Void end

			expr := an_assertion.expression
			if expr /= Void then
				expr.process (Current)
				expr ?= modified_node
				check modified_node /= Void implies expr /= Void end
			end

			create res.make (tag)
			res.set_expression (expr)
			modified_node := res
		end

	process_tagged_indexing (an_indexing: ET_TAGGED_INDEXING) is
			-- Replace tag and terms with their modified form
		local
			tag: ET_TAG
			index: ET_INDEXING
		do
			an_indexing.tag.process (Current)
			tag ?= modified_node
			check tag /= Void end

			process_indexing (an_indexing)
			index ?= modified_node
			check index /= Void end

			create {ET_TAGGED_INDEXING} modified_node.make (tag, index.terms)
		end

	process_true_constant (a_constant: ET_TRUE_CONSTANT) is
			-- Process `a_constant'.
		do
			modified_node := a_constant
		end

	process_tuple_type (a_type: ET_TUPLE_TYPE) is
			-- Replace keyword and actual parameters with their modified form
		local
			key: ET_IDENTIFIER
			params: ET_ACTUAL_PARAMETER_LIST
			res: ET_TUPLE_TYPE
		do
			a_type.tuple_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			params := a_type.actual_parameters
			if params /= Void then
				params.process (Current)
				params ?= modified_node
				check modified_node /= Void implies params /= Void end
			end

			create res.make (params)
			res.set_tuple_keyword (key)
			modified_node := res
		end

	process_type_comma (a_type: ET_TYPE_COMMA) is
			-- Replace type with its modified form
		local
			type: ET_TYPE
		do
			good_type (a_type.type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			create {ET_TYPE_COMMA} modified_node.make (type, a_type.comma)
		end

	process_underscored_integer_constant (a_constant: ET_UNDERSCORED_INTEGER_CONSTANT) is
			-- Replace cast type and sign with their modified form
		local
			type: ET_TARGET_TYPE
			sign: ET_SYMBOL_OPERATOR
			res: ET_UNDERSCORED_INTEGER_CONSTANT
		do
			type ?= good_type (a_constant.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'			

			sign := a_constant.sign
			if sign /= Void then
				sign.process (Current)
				sign ?= modified_node
				check modified_node /= Void implies sign /= Void end
			end

			create res.make (a_constant.literal)
			res.set_cast_type (type)
			res.set_sign (sign)
			modified_node := res
		end

	process_underscored_real_constant (a_constant: ET_UNDERSCORED_REAL_CONSTANT) is
			-- Replace cast type and sign with their modified form
		local
			type: ET_TARGET_TYPE
			sign: ET_SYMBOL_OPERATOR
			res: ET_UNDERSCORED_REAL_CONSTANT
		do
			type ?= good_type (a_constant.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_constant.type'			

			sign := a_constant.sign
			if sign /= Void then
				sign.process (Current)
				sign ?= modified_node
				check modified_node /= Void implies sign /= Void end
			end

			create res.make (a_constant.literal)
			res.set_cast_type (type)
			res.set_sign (sign)
			modified_node := res
		end

	process_unique_attribute (a_feature: ET_UNIQUE_ATTRIBUTE) is
			-- Replace keywords, name, type and assigner with their modified form
		local
			frozen_key, is_key, unique_key: ET_KEYWORD
			type: ET_DECLARED_TYPE
			extended_name: ET_EXTENDED_FEATURE_NAME
			assigner: ET_ASSIGNER
			res: ET_UNIQUE_ATTRIBUTE
		do
			frozen_key := a_feature.frozen_keyword
			if frozen_key /= Void then
				frozen_key.process (Current)
				frozen_key ?= modified_node
				check modified_node /= Void implies frozen_key /= Void end
			end
			a_feature.extended_name.process (Current)
			extended_name ?= modified_node
			check extended_name /= Void end

			good_type (a_feature.declared_type, context, universe).process (Current)
			type ?= modified_node
			check type /= Void end

			assigner := a_feature.assigner
			if assigner /= Void then
				assigner.process (Current)
				assigner ?= modified_node
				check modified_node /= Void implies assigner /= Void end
			end

			a_feature.is_keyword.process (Current)
			is_key ?= modified_node
			check is_key /= Void end

			a_feature.unique_keyword.process (Current)
			unique_key ?= modified_node
			check unique_key /= Void end

			create res.make (extended_name, type, a_feature.implementation_class)
			res.set_frozen_keyword (frozen_key)
			res.set_is_keyword (is_key)
			res.set_unique_keyword (unique_key)
			res.set_assigner (assigner)

			modified_node := res
		end

	process_variant (a_variant: ET_VARIANT) is
			-- Replace keyword, tag and expression with their modified form
		local
			key: ET_KEYWORD
			tag: ET_TAG
			expr: ET_EXPRESSION
			res: ET_VARIANT
		do
			a_variant.variant_keyword.process (Current)
			key ?= modified_node
			check key /= Void end

			tag := a_variant.tag
			if tag /= Void then
				tag.process (Current)
				tag ?= modified_node
				check modified_node /= Void implies tag /= Void end
			end

			expr := a_variant.expression
			if expr /= Void then
				expr.process (Current)
				expr ?= modified_node
				check modified_node /= Void implies expr /= Void end
			end

			create res.make (tag, expr)
			res.set_variant_keyword (key)
			modified_node := res
		end

	process_verbatim_string (a_string: ET_VERBATIM_STRING) is
			-- Replace cast type with its modified form
		local
			type: ET_TARGET_TYPE
			res: ET_VERBATIM_STRING
		do
			type ?= good_type (a_string.cast_type, context, universe)
			if type /= Void then
				type.process (Current)
				type ?= modified_node
				check modified_node /= Void implies type /= Void end
			end
			-- TODO: recalculate `a_string.type'			
			create res.make (a_string.literal, a_string.value, a_string.marker,
				a_string.open_white_characters, a_string.close_white_characters, a_string.is_left_aligned)
			res.set_cast_type (type)
			modified_node := res
		end

	process_void (an_expression: ET_VOID) is
			-- Process `an_expression'.
		do
			modified_node := an_expression
		end

	process_when_part (a_when_part: ET_WHEN_PART) is
			-- Replace choices and compound with their modified form
		local
			choices: ET_CHOICE_LIST
			compound: ET_COMPOUND
		do
			a_when_part.choices.process (Current)
			choices ?= modified_node
			check choices /= Void end

			compound := a_when_part.then_compound
			if compound /= Void then
				compound.process (Current)
				compound ?= modified_node
				check modified_node /= Void implies compound /= Void end
			end

			create {ET_WHEN_PART} modified_node.make (choices, compound)
		end

	process_when_part_list (a_list: ET_WHEN_PART_LIST) is
			-- Replace each when part with its modified form
		local
			i, nb: INTEGER
			item: ET_WHEN_PART
			res: ET_WHEN_PART_LIST
		do
			nb := a_list.count
			create res.make_with_capacity (nb)
			from i := nb until i < 1 loop
				a_list.item (i).process (Current)
				item ?= modified_node
				check item /= Void end
				res.put_first (item)
				i := i - 1
			end
			modified_node := res
		end

feature {NONE} -- Implementation
	context: ET_TYPE_CONTEXT
			-- Current type context
end
